"""
    saraki.erros
    ~~~~~~~~~~~~
"""
from flask import jsonify
from .exc import (
    NotFoundCredentialError,
    InvalidUserError,
    InvalidOrgError,
    InvalidMemberError,
    InvalidPasswordError,
    JWTError,
    TokenNotFoundError,
    AuthorizationError,
    ValidationError,
)


def init_app(app):
    @app.errorhandler(NotFoundCredentialError)
    def not_found_credential_handler(error):
        message = "Provide an access token or a username/password"
        return jsonify({"error": message}), 400

    @app.errorhandler(InvalidUserError)
    def invalid_user_handler(error):
        return jsonify({"error": "Invalid username or password"}), 400

    @app.errorhandler(InvalidPasswordError)
    def invalid_password_handler(error):
        return jsonify({"error": "Invalid username or password"}), 400

    @app.errorhandler(InvalidOrgError)
    def invalid_org_handler(error):
        return jsonify({"error": "Invalid organization"}), 400

    @app.errorhandler(InvalidMemberError)
    def invalid_member_handler(error):
        return jsonify({"error": str(error)}), 400

    @app.errorhandler(JWTError)
    def jwt_error_handler(error):
        return jsonify({"error": str(error)}), 400

    @app.errorhandler(400)
    def bad_request_400(error):
        return jsonify({"error": error.description}), 400

    @app.errorhandler(TokenNotFoundError)
    def not_found_token_handler(error):
        message = "The request does not contain an access token"
        return jsonify({"error": message}), 401

    @app.errorhandler(AuthorizationError)
    def authorization_error_handler(error):
        message = "Invalid access token for this resource"
        return jsonify({"error": message}), 401

    @app.errorhandler(ValidationError)
    def validation_error_handler(e):
        return jsonify({"errors": e.errors}), 400



from json import dumps
from passlib.hash import bcrypt_sha256
from saraki.model import User


def test_signup_user_endpoint(client):

    data = {"email": "elmer@acme", "username": "Elmer", "password": "123"}

    rv = client.post("/signup", data=dumps(data), content_type="application/json")

    assert rv.status_code == 201

    user = User.query.filter_by(username="Elmer").one()

    assert user.email == "elmer@acme"
    assert user.username == "Elmer"
    assert user.canonical_username == "elmer"
    assert bcrypt_sha256.identify(user.password) is True


def test_password_column():
    user = User()
    user.password = "12345"

    assert user.password != "12345"
    assert bcrypt_sha256.identify(user.password) is True


def test_canonical_username_column():
    user = User()
    user.username = "MoMo"

    assert user.canonical_username == "momo"


def test_verify_password():
    user = User()
    user.password = "12345"

    assert user.password != "12345"
    assert user.verify_password("12345") is True



from saraki import Saraki
from saraki.auth import Auth
from saraki.model import database
from flask_sqlalchemy import SQLAlchemy


def test_basic():
    app = Saraki(__name__)

    assert isinstance(app.auth, Auth)
    assert app.extensions["sqlalchemy"].db is database
    assert "app" in app.blueprints


def test_without_auth_object():
    app = Saraki(__name__, auth=None)

    assert not hasattr(app, "auth")
    assert "app" not in app.blueprints


def test_passing_custom_auth_object():
    auth = Auth()
    app = Saraki(__name__, auth=auth)

    assert app.auth is auth
    assert "app" in app.blueprints


def test_without_database_object():
    app = Saraki(__name__, db=None)

    assert "sqlalchemy" not in app.extensions


def test_passing_custom_database_object():
    custom_database = SQLAlchemy()
    app = Saraki(__name__, db=custom_database)

    assert app.extensions["sqlalchemy"].db is custom_database



import pytest
from unittest.mock import MagicMock, patch, call
from flask import Flask, request
from flask.json import dumps
from sqlalchemy import Column, Integer
from sqlalchemy.sql.expression import or_
from sqlalchemy import func, Text

from saraki.auth import current_org
from saraki.endpoints import add_resource, collection
from saraki.model import database, Org, Model
from saraki.exc import ValidationError
from saraki.testing import get_view_function, assert_allowed_methods

from common import Person, Product, OrderLine, Cartoon, Todo, login, auth_ctx


class Test_add_resource:
    def test_basic_usage(self):
        app = Flask(__name__)
        add_resource(app, Cartoon)

        assert_allowed_methods("/cartoon", ["GET", "POST"], app)
        assert_allowed_methods("/cartoon/1", ["GET", "DELETE", "PATCH"], app)

    def test_custom_base_url(self):
        app = Flask(__name__)
        add_resource(app, Cartoon, "animations")

        assert_allowed_methods("/animations", ["GET", "POST"], app)
        assert_allowed_methods("/animations/1", ["GET", "DELETE", "PATCH"], app)

    def test_resource_with_composite_ident(self):
        app = Flask(__name__)
        add_resource(app, OrderLine)

        assert_allowed_methods("/order-line", ["GET", "POST"], app)
        assert_allowed_methods("/order-line/3,14", ["GET", "DELETE", "PATCH"], app)

    def test_resource_with_custom_ident(self):
        app = Flask(__name__)
        add_resource(app, Cartoon, ident="nickname")

        assert_allowed_methods("/cartoon", ["GET", "POST"], app)
        assert_allowed_methods("/cartoon/coyote", ["GET", "DELETE", "PATCH"], app)

    def test_exclude_http_methods(self):
        app = Flask(__name__)
        add_resource(app, Cartoon, methods={"list": ["GET"], "item": ["GET", "PATCH"]})

        assert_allowed_methods("/cartoon", ["GET"], app)
        assert_allowed_methods("/cartoon/1", ["GET", "PATCH"], app)

    def test_endpoint_name(self):
        app = Flask(__name__)
        add_resource(app, Cartoon)
        adapter = app.url_map.bind("")

        list_endpoint = adapter.match("/cartoon")
        add_endpoint = adapter.match("/cartoon", method="POST")
        get_endpoint = adapter.match("/cartoon/1")
        update_endpoint = adapter.match("/cartoon/1", method="PATCH")
        delete_endpoint = adapter.match("/cartoon/1", method="DELETE")

        assert list_endpoint[0] == "list_cartoon"
        assert add_endpoint[0] == "add_cartoon"
        assert get_endpoint[0] == "get_cartoon"
        assert update_endpoint[0] == "update_cartoon"
        assert delete_endpoint[0] == "delete_cartoon"

    def test_custom_resource_name(self):
        app = Flask(__name__)

        add_resource(app, Cartoon, resource_name="film")

        func = get_view_function("/cartoon", app=app)[0]
        assert func._auth_metadata["resource"] == "film"

        func = get_view_function("/cartoon", method="POST", app=app)[0]
        assert func._auth_metadata["resource"] == "film"

        func = get_view_function("/cartoon/1", method="PATCH", app=app)[0]
        assert func._auth_metadata["resource"] == "film"

        func = get_view_function("/cartoon/1", method="DELETE", app=app)[0]
        assert func._auth_metadata["resource"] == "film"

    def test_endpoint_with_same_resource_name(self):
        app = Flask(__name__)

        add_resource(app, Cartoon, resource_name="catalog")
        add_resource(app, Product, resource_name="catalog")

    def test_add_organization_resource(self, app):
        add_resource(app, Todo, "todos")

        assert_allowed_methods("/orgs/acme/todos", ["GET", "POST"], app)
        assert_allowed_methods("/orgs/acme/todos/1", ["GET", "DELETE", "PATCH"], app)


@pytest.mark.usefixtures("data")
class TestResourceList:
    def test_list_endpoint(self, app, client):

        add_resource(app, Cartoon, secure=False)

        rv = client.get("/cartoon")

        assert rv.status_code == 200

        data = rv.get_json()
        assert len(data) == 3


@pytest.mark.usefixtures("data")
class TestGetResourceItem:
    def test_unknown_resource(self, app, client):
        add_resource(app, Cartoon, "cartoons", secure=False)

        rv = client.get("/cartoons/100")
        assert rv.status_code == 404

    def test_get_resource(self, app, client):
        add_resource(app, Cartoon, secure=False)

        rv = client.get("/cartoon/1")
        assert rv.status_code == 200

    def test_resource_with_composite_ident(self, app, client):
        add_resource(app, OrderLine, secure=False)

        rv = client.get("/order-line/1,2")

        assert rv.status_code == 200

        data = rv.get_json()
        assert data["product_id"] == 2

    def test_resource_with_custom_ident(self, app, client):
        add_resource(app, Cartoon, ident="nickname", secure=False)

        rv = client.get("/cartoon/bugs")
        assert rv.status_code == 200

    def test_unknown_resource_with_custom_ident(self, app, client):
        add_resource(app, Cartoon, ident="nickname", secure=False)

        rv = client.get("/cartoon/unknown")
        assert rv.status_code == 404


@pytest.mark.usefixtures("data")
class TestAddResourceItem:
    def test_add_item_endpoint(self, client, app, secure=False):
        add_resource(app, Cartoon, secure=False)

        rv = client.post(
            "/cartoon",
            data=dumps({"name": "Yosemite Sam"}),
            content_type="application/json",
        )
        assert rv.status_code == 201

        data = rv.get_json()
        assert data["name"] == "Yosemite Sam"

    def test_with_invalid_payload(self, app, client):
        add_resource(app, Cartoon, secure=False)

        rv = client.post(
            "/cartoon",
            data=dumps({"unknown": "value"}),
            content_type="application/json",
        )

        assert rv.status_code == 400


@pytest.mark.usefixtures("data")
class TestUpdateResourceItem:
    def test_unknown_resource(self, app, client):
        add_resource(app, Cartoon, "cartoons", secure=False)

        rv = client.patch("/cartoons/100")
        assert rv.status_code == 404

    def test_update(self, app, client):
        add_resource(app, Cartoon, secure=False)

        rv = client.patch(
            "/cartoon/1",
            data=dumps({"name": "Super H."}),
            content_type="application/json",
        )

        data = rv.get_json()

        assert rv.status_code == 200
        assert data["name"] == "Super H."

        rv = client.get("/cartoon/1")
        data = rv.get_json()

        assert rv.status_code == 200
        assert data["name"] == "Super H."

    def test_invalid_payload(self, app, client):
        add_resource(app, Cartoon, secure=False)

        rv = client.patch(
            "/cartoon/1",
            data=dumps({"unknown": "Fried chicken"}),
            content_type="application/json",
        )

        assert rv.status_code == 400


@pytest.mark.usefixtures("data")
class TestDeleteResourceItem:
    def test_unknown_resource(self, app, client):
        add_resource(app, Cartoon, "cartoons", secure=False)

        rv = client.delete("/cartoons/100")
        assert rv.status_code == 404

    def test_delete(self, app, client):
        add_resource(app, Cartoon, "cartoons", secure=False)

        rv = client.delete("/cartoons/1")
        assert rv.status_code == 200

        rv = client.get("/cartoons/1")
        assert rv.status_code == 404


@pytest.mark.usefixtures("data")
class TestResourceAuthorization:
    def test_get_resource_list(self, app, client):
        add_resource(app, Cartoon)
        token = login("coyote", scope={"cartoon": ["read"]})

        rv = client.get("/cartoon", headers={"Authorization": token})
        assert rv.status_code != 401

    def test_get_resource_item(self, app, client):
        add_resource(app, Cartoon)
        token = login("coyote", scope={"cartoon": ["read"]})

        rv = client.get("/cartoon/1", headers={"Authorization": token})
        assert rv.status_code != 401

    def test_add_resource_item(self, app, client):
        add_resource(app, Cartoon)
        token = login("coyote", scope={"cartoon": ["write"]})

        rv = client.post("/cartoon", headers={"Authorization": token})
        assert rv.status_code != 401

    def test_update_resource_item(self, app, client):
        add_resource(app, Cartoon)
        token = login("coyote", scope={"cartoon": ["write"]})

        rv = client.post("/cartoon", headers={"Authorization": token})
        assert rv.status_code != 401

    def test_delete_resource_item(self, app, client):
        add_resource(app, Cartoon)
        token = login("coyote", scope={"cartoon": ["delete"]})

        rv = client.delete("/cartoon", headers={"Authorization": token})
        assert rv.status_code != 401


@pytest.mark.usefixtures("data", "data_org")
class TestOrgResource:
    def insert_data(self):
        org_id = Org.query.filter_by(orgname="acme").one().id

        todo = Todo(task="Stop being lazy", org_id=org_id)
        database.session.add(todo)
        database.session.commit()

        return todo

    def test_list_resource(self, app, client):
        _id = self.insert_data().id
        add_resource(app, Todo)

        # Request to Acme endpoint
        token = login("coyote", "acme", scope={"todo": ["read"]})
        rv = client.get(f"/orgs/acme/todo", headers={"Authorization": token})
        data = rv.get_json()

        assert rv.status_code == 200
        assert data[0]["id"] == _id
        assert "org_id" not in data[0]

        # Request to R.R. Inc endpoint
        token = login("RoadRunner", "rrinc", scope={"todo": ["read"]})
        rv = client.get(f"/orgs/rrinc/todo", headers={"Authorization": token})
        data = rv.get_json()

        assert rv.status_code == 200
        assert len(data) == 0

    def test_get_resource(self, app, client):
        _id = self.insert_data().id
        add_resource(app, Todo)

        # Request to R.R. Inc endpoint
        token = login("RoadRunner", "rrinc", scope={"todo": ["read"]})
        rv = client.get(f"/orgs/rrinc/todo/{_id}", headers={"Authorization": token})

        assert rv.status_code == 404, (
            f"The status code should be 404 because the model id ({_id})"
            " used in the URL belongs to another organization account."
        )

        # Request to Acme endpoint
        token = login("coyote", "acme", scope={"todo": ["read"]})
        rv = client.get(f"/orgs/acme/todo/{_id}", headers={"Authorization": token})
        data = rv.get_json()

        assert rv.status_code == 200
        assert data["id"] == _id
        assert data["task"] == "Stop being lazy"
        assert "org_id" not in data

    def test_add_resource(self, app, client):
        add_resource(app, Todo)

        token = login("coyote", "acme", scope={"todo": ["write"]})

        rv = client.post(
            "/orgs/acme/todo",
            data=dumps({"task": "Do something"}),
            content_type="application/json",
            headers={"Authorization": token},
        )
        data = rv.get_json()

        assert rv.status_code == 201
        assert "org_id" not in data

        org = Org.query.filter_by(orgname="acme").one()
        todo = Todo.query.get(data["id"])

        assert org.id == todo.org_id

    def test_update_resource(self, app, client):
        _id = self.insert_data().id
        add_resource(app, Todo)

        # Request to R.R. Inc endpoint with the same id
        token = login("RoadRunner", "rrinc", scope={"todo": ["write"]})
        rv = client.patch(
            f"/orgs/rrinc/todo/{_id}",
            data=dumps({"task": "Do something"}),
            content_type="application/json",
            headers={"Authorization": token},
        )
        assert rv.status_code == 404, (
            f"The status code should be 404 because the model id ({_id})"
            " used in the URL belongs to another organization account."
        )

        # Request to Acme endpoint
        token = login("coyote", "acme", scope={"todo": ["write"]})
        rv = client.patch(
            f"/orgs/acme/todo/{_id}",
            data=dumps({"task": "Do something"}),
            content_type="application/json",
            headers={"Authorization": token},
        )
        data = rv.get_json()

        assert rv.status_code == 200
        assert "org_id" not in data

        todo = Todo.query.filter_by(id=_id).one()
        assert todo.task == "Do something"

    def test_delete_resource(self, app, client):
        _id = self.insert_data().id
        add_resource(app, Todo)

        # Request to R.R. Inc.
        token = login("RoadRunner", "rrinc", scope={"todo": ["delete"]})
        headers = {"Authorization": token}
        rv = client.delete(f"/orgs/rrinc/todo/{_id}", headers=headers)

        assert rv.status_code == 404, (
            f"The status code should be 404 because the model id ({_id})"
            " used in the URL belongs to another organization account."
        )

        # Request to Acme
        token = login("coyote", "acme", scope={"todo": ["delete"]})
        headers = {"Authorization": token}
        rv = client.delete(f"/orgs/acme/todo/{_id}", headers=headers)

        assert rv.status_code == 200
        assert Todo.query.get(_id) is None


def mock_query():
    query = MagicMock(
        name="Query", spec=["filter", "filter_by", "order_by", "paginate"]
    )
    query.filter_by = MagicMock(return_value=query)
    query.order_by = MagicMock(return_value=query)
    query.filter = MagicMock(return_value=query)

    result = MagicMock(name="result")
    result.items = []

    query.paginate = MagicMock(return_value=result)

    return query


@pytest.mark.usefixtures("database_conn", "ctx")
class TestCollection:
    @pytest.mark.parametrize(
        "key, value, expected",
        [
            (
                "search",
                '{"t": "term", "f": ["name", "color"]}',
                {"t": "term", "f": ["name", "color"]},
            ),
            ("filter", '{"color": "blue", "price": 2}', {"color": "blue", "price": 2}),
            ("select", '{"id": 1, "name": 1}', {"id": 1, "name": 1}),
            ("limit", "50", 50),
            ("page", "3", 3),
        ],
    )
    def test_parse_query_string(self, request_ctx, key, value, expected):
        with request_ctx(f"/?{key}={value}"):
            output = collection._parse_query_string(Product, request.args)

        assert output[key] == expected

    @pytest.mark.parametrize(
        "qs",
        [
            'filter={"unknown": "value", "price": 20}',
            'filter={"unknown": "value", "price": "text"}',
            "filter={name: tnt, enabled: True}",
            'select={"unknown": 1, "name": 1}',
            'select={"id": "text", "name": 1}',
            "select={id: 1, name: 1}",
            "limit=text",
            "page=text",
        ],
        ids=[
            "filter: with unknown column",
            "filter: with invalid value type",
            "filter: invalid JSON string",
            "select: with unknown column",
            "select: with non integer value",
            "select: invalid JSON string",
            "limit : non integer value",
            "page  : non integer value",
        ],
    )
    def test_invalid_modifier(self, request_ctx, qs):
        with request_ctx(f"/?{qs}"):
            with pytest.raises(ValidationError):
                collection()(lambda: Product)()

    def test_unknown_modifier(self, request_ctx):
        with request_ctx("/?unknown=value"):
            with pytest.raises(ValidationError):
                collection()(lambda: Product)()

    @pytest.mark.parametrize(
        "param, expected",
        [
            ({"name": 1, "color": 1}, {"include": ["name", "color"]}),
            ({"name": 1, "color": 0}, {"include": ["name"], "exclude": ["color"]}),
            ({"name": 0, "color": 0}, {"exclude": ["name", "color"]}),
        ],
    )
    @patch.object(Product, "query")
    def test_parse_select_modifier(self, query, request_ctx, param, expected):
        params = collection._parse_select_modifier(param)

        assert params == expected

    @patch.object(Product, "query", new=mock_query())
    def test_filter_modifier(self, request_ctx):

        with request_ctx('/?filter={"name": "tnt", "enabled": true}'):
            collection()(lambda: Product)()

        Product.query.filter_by.assert_called_with(**{"name": "tnt", "enabled": True})

    @patch.object(Product, "query", new=mock_query())
    def test_search_modifier(self, request_ctx):

        with request_ctx('/?search={"t": "black", "f": ["name", "color"]}'):
            collection()(lambda: Product)()

        Product.query.filter.assert_called()

        expected = or_(Product.name.ilike("%black%"), Product.color.ilike("%black%"))
        original = Product.query.filter.call_args[0][0]

        assert original.compare(expected)

    @patch.object(Product, "query", new=mock_query())
    def test_search_modifier_on_non_text_column(self, request_ctx):

        with request_ctx('/?search={"t": "black", "f": ["price", "color"]}'):
            collection()(lambda: Product)()

        Product.query.filter.assert_called()

        expected = or_(
            func.cast(Product.price, Text).ilike("%black%"),
            Product.color.ilike("%black%"),
        )
        original = Product.query.filter.call_args[0][0]

        # clauseelement.compare() is not implemented for cast()
        # so use string comparison instead
        assert str(original) == str(expected)

    @patch.object(Product, "query", new=mock_query())
    def test_sort_modifier_with_single_field(self, request_ctx):
        with request_ctx("/?sort=name"):
            collection()(lambda: Product)()

        Product.query.order_by.assert_called()

        expected = Product.name.asc()
        original = Product.query.order_by.call_args[0][0]
        assert original.compare(expected)

    @patch.object(Product, "query", new=mock_query())
    def test_sort_modifier_with_multiple_fields(self, request_ctx):
        with request_ctx("/?sort=name,color"):
            collection()(lambda: Product)()

        Product.query.order_by.assert_called()

        expected = Product.name.asc()
        original = Product.query.order_by.call_args[0][0]
        assert original.compare(expected)

        expected = Product.color.asc()
        original = Product.query.order_by.call_args[0][1]
        assert original.compare(expected)

    @patch.object(Product, "query", new=mock_query())
    def test_sort_modifier_with_descending_order(self, request_ctx):
        with request_ctx("/?sort=-name"):
            collection()(lambda: Product)()

        Product.query.order_by.assert_called()

        expected = Product.name.desc()
        original = Product.query.order_by.call_args[0][0]
        assert original.compare(expected)

    # fmt: off
    @pytest.mark.parametrize(
        "query_string, expected",
        [
            ("", (1, 30)),
            ("limit=50", (1, 50)),
            ("limit=200", (1, 100))
        ],
        ids=[
            "default limit",
            "custom limit",
            "default max limit"
        ],
    )
    @patch.object(Product, "query", new=mock_query())
    def test_limit_modifier(self, request_ctx, query_string, expected):
        # The mock is created just one time, so reset it
        Product.query.paginate.reset_mock()

        with request_ctx(f"/?{query_string}"):
            collection()(lambda: Product)()

        Product.query.paginate.assert_called_with(*expected)
    # fmt: on

    @patch.object(Product, "query", new=mock_query())
    def test_page_modifier(self, request_ctx):
        with request_ctx("/?page=50"):
            collection()(lambda: Product)()

        Product.query.paginate.assert_called_with(50, 30)

    @patch.object(Product, "query", new=mock_query())
    def test_select_modifier(self, request_ctx):
        item = MagicMock()
        Product.query.paginate().items = [item]

        with request_ctx('/?select={"id": 1}'):
            collection()(lambda: Product)()

        item.export_data.assert_called_with(["id"], ())

    @patch.object(Product, "query")
    def test_returned_value(self, query, request_ctx):
        item = MagicMock()
        item.export_data.return_value = {"id": 300}

        query.paginate().items = [item]
        query.paginate().total = 1

        with request_ctx('/?select={"id": 1}&page=3'):
            rv = collection()(lambda: Product)()

        assert rv[0] == [{"id": 300}]
        assert rv[1] == {"X-Total": 1, "X-Page": 3}

    @patch.object(Product, "query")
    def test_model_class_without_export_data_method(self, query, request_ctx):
        item = Product(id=4, name="Acme explosive tennis balls")

        query.paginate().items = [item]
        query.paginate().total = 1

        with request_ctx('/?select={"id": 1}&page=3'):
            rv = collection()(lambda: Product)()

        assert rv[0] == [{"id": 4}]
        assert rv[1] == {"X-Total": 1, "X-Page": 3}

    @patch.object(Person, "query")
    def test_model_class_with_export_data_method(self, query, request_ctx):
        item = Person(id=300, firstname="John", lastname="Connor")

        query.paginate().items = [item]
        query.paginate().total = 1

        with request_ctx('/?select={"firstname": 1}&page=3'):
            rv = collection()(lambda: Person)()

        assert rv[0] == [{"firstname": "John"}]
        assert rv[1] == {"X-Total": 1, "X-Page": 3}

    def test_model_class_export_data_with_exception(self, request_ctx):
        class TestTable(Model):
            id = Column(Integer, primary_key=True)

            def export_data(self, *args, **kargs):
                raise AttributeError("Inside of export_data method")

        TestTable.query = MagicMock()
        TestTable.query.paginate().items = [TestTable(id=1)]

        with request_ctx("/"):
            with pytest.raises(AttributeError, match="Inside of export_data method"):
                collection()(lambda: TestTable)()

    @patch.object(Todo, "query", new=mock_query())
    def test_basic_request_with_org_model(self, app):

        with app.test_request_context("/"):
            with auth_ctx("coyote", "acme"):
                current_org_id = current_org.id
                collection()(lambda: Todo)()

        Todo.query.filter_by.assert_called_with(org_id=current_org_id)

    @patch.object(Todo, "query", new=mock_query())
    def test_filter_modifier_with_org_model(self, request_ctx):

        with request_ctx('/?filter={"task": "Do something"}'):
            with auth_ctx("coyote", "acme"):
                current_org_id = current_org.id
                collection()(lambda: Todo)()

        Todo.query.filter_by.assert_called_with(
            org_id=current_org_id, task="Do something"
        )

    @patch.object(Todo, "query", new=mock_query())
    def test_search_modifier_with_org_model(self, request_ctx):

        with request_ctx('/?search={"t": "black", "f": ["task"]}'):
            with auth_ctx("coyote", "acme"):
                current_org_id = current_org.id
                collection()(lambda: Todo)()

        Todo.query.filter_by.assert_called_with(org_id=current_org_id)
        Todo.query.filter.assert_called()

        expected = or_(Todo.task.ilike("%black%"))
        original = Todo.query.filter.call_args[0][0]
        assert original.compare(expected)

        # The query should be filtered by the organization id first
        assert Todo.query.method_calls[0] == call.filter_by(org_id=current_org_id)



class ProgrammingError(Exception):
    pass


class AuthenticationError(Exception):
    pass


class InvalidUserError(AuthenticationError):
    pass


class InvalidOrgError(AuthenticationError):
    pass


class InvalidPasswordError(AuthenticationError):
    pass


class InvalidMemberError(AuthenticationError):
    pass


class NotFoundCredentialError(AuthenticationError):
    """Raised when a token or a username/password pair can not be found in
    the current HTTP request."""

    pass


class TokenNotFoundError(AuthenticationError):
    pass


class JWTError(AuthenticationError):
    pass


class AuthorizationError(Exception):
    pass


class ValidationError(Exception):
    def __init__(self, errors):
        self.errors = errors



import os
import pytest

from saraki import Saraki
from saraki.model import database
from saraki.testing import Savepoint

from data import (
    insert_cartoons,
    insert_persons,
    insert_products,
    insert_orders,
    insert_actions,
    insert_resources,
    insert_plans,
    insert_users,
    insert_orgs,
    insert_members,
    insert_roles,
    insert_member_roles,
)


from assertions import pytest_assertrepr_compare  # noqa: F401


needdatabase = pytest.mark.skipif(
    os.getenv("DATABASE_URI") is None,
    reason="This need a database connection and DATABASE_URI is not defined",
)


@pytest.fixture(scope="session")
def _setup_database(request):
    """Setup the database

    Creates all tables on setup and deletes all tables on cleanup.
    """

    _app = Saraki(__name__, db=None)
    _app.config["SQLALCHEMY_DATABASE_URI"] = os.environ["TEST_DATABASE_URI"]
    database.init_app(_app)

    with _app.app_context():
        database.create_all()

    def teardown():
        with _app.app_context():
            database.session.remove()
            database.drop_all()

    request.addfinalizer(teardown)

    return database


@pytest.fixture(scope="session")
def _insert_data(_setup_database):
    """Put the database in a known state by inserting predefined data."""

    _app = Saraki(__name__, db=None)
    _app.config["SQLALCHEMY_DATABASE_URI"] = os.environ["TEST_DATABASE_URI"]
    database.init_app(_app)

    with _app.app_context():
        insert_actions()

    with _app.app_context():
        insert_resources()

    with _app.app_context():
        insert_persons()
        insert_products()
        insert_orders()
        insert_cartoons()
        insert_plans()
        insert_users()

        # Insert all registered resources and actions
        _app.init()

        database.session.commit()


@pytest.fixture
def data(_insert_data, database_conn):
    pass


@pytest.fixture
def data_org(ctx, savepoint):
    insert_orgs()


@pytest.fixture
def data_member(ctx, savepoint, data_org):
    insert_members()


@pytest.fixture
def data_role(ctx, savepoint, data_org):
    insert_roles()


@pytest.fixture
def data_member_role(ctx, savepoint, data_member, data_role):
    insert_member_roles()


@pytest.fixture(scope="session")
def _trn():
    """Create a session wide instance of Savepoint class.

    Savepoint helps create nested database transactions.
    This help rollback any transaction using PostgreSQL savepoints.
    """
    return Savepoint(database)


@pytest.fixture
def app(request):

    app = Saraki("flask_test", root_path=os.path.dirname(__file__), db=None)
    app.config["TESTING"] = True
    app.config["SECRET_KEY"] = "secret"

    return app


@pytest.fixture
def request_ctx(app):
    return app.test_request_context


@pytest.fixture
def database_conn(app):
    app.config["SQLALCHEMY_DATABASE_URI"] = os.environ.get("TEST_DATABASE_URI")
    database.init_app(app)


@pytest.fixture
def savepoint(_setup_database, database_conn, _trn, request):
    _trn.start()

    def teardown():
        _trn.end()

    request.addfinalizer(teardown)


@pytest.fixture
def ctx(app, request):
    """Push a new application context and closes it automatically
    when the test goes out of scope.

    This helps to avoid creating application contexts manually
    either by calling app.app_context() with python `with` statement or
    by pushing or popping manually.
    """
    ctx = app.app_context()
    ctx.push()

    def teardown():
        ctx.pop()

    request.addfinalizer(teardown)

    return ctx


@pytest.fixture
def client(app, ctx, savepoint, database_conn):
    """Flask Test Client

    This starts a database nested transaction and then closes it
    when the test goes out of scope in order to rollback any change
    made to the database.
    """

    return app.test_client()


@pytest.fixture
def xclient(app):
    """Flask Test Client (No DB transaction)"""

    return app.test_client()



import pytest
from json import loads
from flask import abort

from saraki.exc import (
    NotFoundCredentialError,
    InvalidUserError,
    InvalidPasswordError,
    InvalidOrgError,
    InvalidMemberError,
    JWTError,
    TokenNotFoundError,
    AuthorizationError,
)

parametrize = pytest.mark.parametrize


def test_not_found_credential_handler(app, client):
    @app.route("/")
    def index():
        raise NotFoundCredentialError()

    rv = client.get("/")
    data = loads(rv.data)

    assert rv.status_code == 400
    assert data["error"] == "Provide an access token or a username/password"


def test_invalid_user_handler(app, client):
    @app.route("/")
    def index():
        raise InvalidUserError()

    client = app.test_client()
    rv = client.get("/")
    data = loads(rv.data)

    assert rv.status_code == 400
    assert data["error"] == "Invalid username or password"


def test_invalid_password_handler(app, client):
    @app.route("/")
    def index():
        raise InvalidPasswordError()

    rv = client.get("/")
    data = loads(rv.data)

    assert rv.status_code == 400
    assert data["error"] == "Invalid username or password"


def test_invalid_org_handler(app, client):
    @app.route("/")
    def index():
        raise InvalidOrgError()

    rv = client.get("/")
    data = loads(rv.data)

    assert rv.status_code == 400
    assert data["error"] == "Invalid organization"


def test_invalid_member_handler(app, client):
    @app.route("/")
    def index():
        raise InvalidMemberError("Invalid member message")

    rv = client.get("/")
    data = loads(rv.data)

    assert rv.status_code == 400
    assert data["error"] == "Invalid member message"


def test_jwt_error_handler(app, client):
    @app.route("/")
    def index():
        raise JWTError("Token related error message")

    rv = client.get("/")
    data = loads(rv.data)

    assert rv.status_code == 400
    assert data["error"] == "Token related error message"


@parametrize("error_value", ("An error message", {"prop": "value"}, 1))
def test_bad_request(app, client, error_value):
    @app.route("/")
    def index():
        abort(400, error_value)

    rv = client.get("/")
    data = loads(rv.data)

    assert rv.status_code == 400
    assert data["error"] == error_value


def test_not_found_token_handler(app, client):
    @app.route("/")
    def index():
        raise TokenNotFoundError()

    rv = client.get("/")
    data = loads(rv.data)

    assert rv.status_code == 401
    assert data["error"] == "The request does not contain an access token"


def test_authorization_error_handler(app, client):
    @app.route("/")
    def index():
        raise AuthorizationError()

    rv = client.get("/")
    data = loads(rv.data)

    assert rv.status_code == 401
    assert data["error"] == "Invalid access token for this resource"



from functools import wraps
from json.decoder import JSONDecodeError

from flask import request, abort
from flask.json import loads as json_loads

from sqlalchemy import inspect
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy import func, Text
from sqlalchemy.sql.expression import or_

from saraki.model import database
from saraki.exc import ValidationError
from saraki.auth import require_auth, current_org
from saraki.utility import (
    Validator,
    json,
    generate_schema,
    export_from_sqla_object as export_data,
    import_into_sqla_object as import_data,
)


class Collection:
    """ Creates a decorator for collection endpoints.

    View functions decorated with this decorator must return a SQLAlchemy
    declarative class. This decorator can handle filtering, search, pagination
    and sorting using HTTP query strings.

    This is implemented as a class to extend or change the format of the query
    strings.

    Usage:

    .. code-block:: python

        # First create a instance
        collection = Collection()

        @app.route('/products')
        @collection()
        def index():
            # return a SQLAlchemy declarative class
            return Product

    """

    def _parse_query_string(self, cls, qs):
        qs = qs.to_dict(flat=True)
        schema = generate_schema(cls)
        mapper = inspect(cls)
        columns = [column.name for column in mapper.c]
        query_string_schema = {
            "select": {
                "type": "dict",
                "allowed": columns,
                "valueschema": {"type": "integer"},
            },
            "search": {
                "type": "dict",
                "schema": {
                    "t": {"required": True},
                    "f": {"required": True, "type": "list", "allowed": columns},
                },
            },
            "filter": {"type": "dict", "schema": schema},
            "sort": {},
            "limit": {"type": "integer", "coerce": int},
            "page": {"type": "integer", "coerce": int},
        }

        # First decode all modifiers with JSON string
        json_keys = {"select", "search", "filter"}
        decoded_qs = {}

        for key, value in qs.items():
            if key in json_keys:
                try:
                    decoded_qs[key] = json_loads(value)
                except JSONDecodeError:
                    raise ValidationError({key: "Invalid JSON string"})
            else:
                decoded_qs[key] = value

        v = Validator(query_string_schema)

        if v.validate(decoded_qs, update=True) is False:
            raise ValidationError(v.errors)

        return v.normalized(decoded_qs)

    def _filter_modifier(self, query, filters):
        return query.filter_by(**filters)

    def _parse_select_modifier(self, select):
        include = []
        exclude = []

        for column_name, flag in select.items():
            if flag:
                include.append(column_name)
                continue

            exclude.append(column_name)

        params = {}

        if include:
            params["include"] = include

        if exclude:
            params["exclude"] = exclude

        return params

    def _search_modifier(self, cls, query, search):
        term = search["t"]
        filters = []
        mapper = cls.__mapper__

        for column_name in search["f"]:
            column = getattr(cls, column_name)

            if mapper.c[column_name].type.python_type != str:
                column = func.cast(column, Text)

            filters.append(column.ilike(f"%{term}%"))

        return query.filter(or_(*filters))

    def _sort_modifier(self, cls, query, sort):
        sorting = []

        for column_name in sort.split(","):
            if column_name.startswith("-"):
                column = getattr(cls, column_name[1:])
                sorting.append(column.desc())
                continue

            column = getattr(cls, column_name)
            sorting.append(column.asc())

        return query.order_by(*sorting)

    def __call__(self, default_limit=30, max_limit=100):
        def decorator(f):
            @wraps(f)
            def wrapper(*args, **kwargs):

                Model = f(*args, **kwargs)
                filters = {}

                if hasattr(Model, "org_id"):
                    filters = {"org_id": current_org.id}

                modifiers = self._parse_query_string(Model, request.args)

                query = Model.query

                if "filter" in modifiers:
                    filters.update(modifiers.get("filter", {}))

                if filters:
                    query = self._filter_modifier(query, filters)

                if "search" in modifiers:
                    query = self._search_modifier(Model, query, modifiers["search"])

                if "sort" in modifiers:
                    query = self._sort_modifier(Model, query, modifiers["sort"])

                page = modifiers.get("page", 1)
                limit = min(modifiers.get("limit", default_limit), max_limit)

                result = query.paginate(page, limit)
                items = result.items

                export_data_params = {}

                if "select" in modifiers:
                    export_data_params = self._parse_select_modifier(
                        modifiers["select"]
                    )

                payload = export_data(items, **export_data_params)

                return payload, {"X-Total": result.total, "X-Page": page}

            return wrapper

        return decorator


#: collection decorator
collection = Collection()


def _import_data(model, data):
    # Classes with the import_data method can customize the import
    # process, therefore, prioritize it.
    if hasattr(model, "import_data"):
        model.import_data(data)
    else:
        import_data(model, data)


def list_view_func(modelcls, ident_prop, primary_key, schema, is_org, **kargs):
    return modelcls


def add_view_func(modelcls, ident_prop, primary_key, schema, is_org, **kargs):
    payload = request.get_json()

    if is_org:
        payload["org_id"] = current_org.id

    v = Validator(schema)

    if v.validate(payload) is False:
        raise ValidationError(v.errors)

    model = modelcls()
    data = v.normalized(payload)
    _import_data(model, data)

    database.session.add(model)
    database.session.commit()

    return model, 201


def item_view(modelcls, ident_prop, primary_key, schema, is_org, **kargs):
    """Generic view function to handle operations on single resource items."""

    ident = {prop: kargs.get(prop) for prop in ident_prop}

    if is_org:
        ident["org_id"] = current_org.id

    try:
        model = modelcls.query.filter_by(**ident).one()
    except NoResultFound:
        abort(404)

    if request.method == "GET":
        return model

    if request.method == "DELETE":
        database.session.delete(model)
        database.session.commit()
        return model

    if request.method == "PATCH":
        payload = request.get_json()
        v = Validator(schema)

        if v.validate(payload, update=True, model=model) is False:
            raise ValidationError(v.errors)

        data = v.normalized(payload)
        _import_data(model, data)
        database.session.commit()

        return model


type_mapping = {int: "int", str: "string"}


def _generate_route_rules(base_url, modelcls, ident_prop, is_org=False):
    list_rule = f"/orgs/<aud:orgname>/{base_url}" if is_org else f"/{base_url}"
    item_rule = f"{list_rule}/"

    columns = [getattr(modelcls, column_name) for column_name in ident_prop]

    for column in columns:
        python_type = column.type.python_type
        _type = type_mapping.get(python_type, "string")
        item_rule += f"<{_type}:{column.name}>,"

    # Remove the last , character
    item_rule = item_rule[:-1]

    return (list_rule, item_rule)


def add_resource(
    app,
    modelcls,
    base_url=None,
    ident=None,
    methods=None,
    secure=True,
    resource_name=None,
):
    """ Register a resource and generate API endpoints to interact with it.

    The first parameter is a SQLAlchemy model class and the second can
    be a Flask app instance or a Blueprint instance.

    Let start with a code example::

        class Product(Model):
            __tablename__ = 'product'

            id = Column(Integer, primary_key=True)
            name = Column(String)

        add_resource(Product, app)

    The above code will generate the next route rules.

    +-----------------------+--------+----------------------------+
    | Route rule            | Method | Description                |
    +=======================+========+============================+
    | ``/product``          | GET    | Retrive a collection       |
    +-----------------------+--------+----------------------------+
    | ``/product``          | POST   | Create a new resource item |
    +-----------------------+--------+----------------------------+
    | ``/product/<int:id>`` | GET    | Retrieve a resource item   |
    +-----------------------+--------+----------------------------+
    | ``/product/<int:id>`` | PATCH  | Update a resource item     |
    +-----------------------+--------+----------------------------+
    | ``/product/<int:id>`` | DELETE | Delete a resource item     |
    +-----------------------+--------+----------------------------+

    By default, the **name** of the table is used to render the resource list
    part of the url and the name of the **primary key** column for the resource
    identifier part. Note that the type of the column is used when possible for
    the route rule variable type.

    If the model class has a composite primary key, the identifier part
    are rendered with each column name separated by a comma.

    For example::

        class OrderLine(Model):
            __tablename__ = 'order_line'

            order_id = Column(Integer, primary_key=True)
            product_id = Column(Integer, primary_key=True)

        add_resource(Product, app)

    The route rules will be::

        /order-line
        /order-line/<int:order_id>,<int:product_id>

    Note that the character (_) was sustituted by a dash (-) character in the
    base url.

    To customize the base url (resource list part) use the ``base_url``
    parameter, for example::

        add_resource(ap, Product, 'products')

    The rendered route rules will be::

        /products
        /products/<int:id>

    By default, all endpoints are secured using the :meth:`~saraki.auth.require_auth`
    decorator. Once again, the table name is used for the resource parameter of
    require_auth, unless the resource_name parameter are provided.

    To disable this behavior pass ``secure=False``.

    Model classes with a property (column) named ``org_id`` will be considered
    an organization resource and will generate an organization endpoint. For
    instance, supposing the model class Product has the property org_id the
    generated route rules will be::

        /orgs/<aud:orgname>/products
        /orgs/<aud:orgname>/products/<int:id>

    :param app: Flask or Blueprint instance.
    :param modelcls: SQLAlchemy model class.
    :param base_url: The base url for the resource.
    :param ident: Names of the column used to identify a resource item.
    :param methods: Dict object with allowd HTTP methods for item and list resources.
    :param secure: Boolean flag to secure a resource using require_auth.
    :param resource_name: resource name required in token scope to access this resource.
    """

    is_org = hasattr(modelcls, "org_id")

    # The table name is used to generate flask endpoint names
    table_name = modelcls.__tablename__

    # When resource_name is not provided use the name of the table
    resource_name = resource_name or table_name

    # Use the table name as default base URL
    base_url = base_url or "-".join(table_name.split("_"))
    primary_key = inspect(modelcls).primary_key
    schema = generate_schema(modelcls)

    primary_key = tuple(column.name for column in primary_key)
    ident = (ident,) if ident else primary_key

    list_rule, item_rule = _generate_route_rules(base_url, modelcls, ident, is_org)

    methods = methods or {}

    list_methods = methods.get("list", ("GET", "POST"))
    item_methods = methods.get("item", ("GET", "PATCH", "DELETE"))

    defaults = {
        "schema": schema,
        "modelcls": modelcls,
        "ident_prop": ident,
        "primary_key": primary_key,
        "is_org": is_org,
    }

    # Resource list
    if "GET" in list_methods:
        endpoint = f"list_{table_name}"
        view_func = json(collection()(list_view_func))

        if secure:
            view_func = require_auth(resource_name)(view_func)

        app.add_url_rule(list_rule, endpoint, view_func, defaults=defaults)

    if "POST" in list_methods:
        endpoint = f"add_{table_name}"
        view_func = json(add_view_func)

        if secure:
            view_func = require_auth(resource_name)(view_func)

        app.add_url_rule(
            list_rule, endpoint, view_func, defaults=defaults, methods=["POST"]
        )

    # Resource item
    view_func = json(item_view)

    if secure:
        view_func = require_auth(resource_name)(view_func)

    if "GET" in item_methods:
        endpoint = f"get_{table_name}"
        app.add_url_rule(
            item_rule, endpoint, view_func, defaults=defaults, methods=["GET"]
        )

    if "PATCH" in item_methods:
        endpoint = f"update_{table_name}"
        app.add_url_rule(
            item_rule, endpoint, view_func, defaults=defaults, methods=["PATCH"]
        )

    if "DELETE" in item_methods:
        endpoint = f"delete_{table_name}"
        app.add_url_rule(
            item_rule, endpoint, view_func, defaults=defaults, methods=["DELETE"]
        )



from flask import Flask, Blueprint as _Blueprint
from flask_sqlalchemy import SQLAlchemy
from saraki import errors
from saraki import config as _default_config
from saraki.model import database
from saraki.auth import Auth
from saraki.endpoints import add_resource


class Saraki(Flask):
    def __init__(self, import_name, auth=Auth, db=database, **kargs):

        super(Saraki, self).__init__(import_name, **kargs)

        self.config.from_object(_default_config)
        self.add_default_endpoints()

        if auth:
            self.auth = auth() if callable(auth) else auth
            self.auth.init_app(self)
            self.add_default_blueprints()

        if isinstance(db, SQLAlchemy):
            db.init_app(self)

        errors.init_app(self)

    def init(self):
        if hasattr(self, "auth"):
            self.auth.persist_data()

        database.session.commit()

    def add_default_blueprints(self):
        from saraki.api import appbp

        self.register_blueprint(appbp)

    def add_default_endpoints(self):
        from saraki.api import signup_view

        self.add_url_rule("/signup", "signup", signup_view, methods=["POST"])

    def add_resource(self, modelcls, base_url=None, **options):
        add_resource(self, modelcls, base_url=base_url, **options)


class Blueprint(_Blueprint):
    def add_resource(self, modelcls, base_url=None, **options):
        add_resource(self, modelcls, base_url=base_url, **options)



import datetime
from functools import wraps

from cerberus import Validator as _Validator
from sqlalchemy import inspect
from sqlalchemy.orm.collections import InstrumentedList
from sqlalchemy.exc import NoInspectionAvailable
from flask import request, abort, jsonify
from flask.wrappers import Response


schema_type_conversions = {
    int: "integer",
    str: "string",
    bool: "boolean",
    datetime.date: "string",
    datetime.datetime: "string",
}


def is_sqla_obj(obj):
    """Checks if an object is a SQLAlchemy model instance."""
    try:
        inspect(obj)
        return True
    except NoInspectionAvailable:
        return False


def import_into_sqla_object(model_instance, data):
    """Import a dictionary, assigning each item value that match to a
    column name of the object entity class.
    """

    mapper = inspect(model_instance.__class__)

    for key in data:
        if key in mapper.c:
            setattr(model_instance, key, data[key])

    return model_instance


def _get_column_default(c):
    d = c.default
    return d.arg if isinstance(getattr(d, "arg", None), (int, str, bool)) else None


class ExportData:
    """Creates a callable object that convert SQLAlchemy model instances
    to dictionaries.
    """

    def __init__(self, exclude=()):

        #: A global list of column names to exclude. This takes precedence over
        #: the parameters ``include`` and/or ``exclude`` of this instance call.
        self.exclude = tuple(exclude)

    def __call__(self, obj, include=(), exclude=()):
        """Converts SQLAlchemy models into python serializable objects. It can
        take a single model or a list of models.

        By default, all columns are included in the output, unless a list of
        column names are provided to the parameters ``include`` or ``exclude``.
        The latter has precedence over the former. Finally, the columns that
        appear in the :attr:`excluded` property will be excluded, regardless of
        the values that the parameters include and exclude have.

        If the model is not persisted in the database, the default values of
        the columns are used if they exist in the class definition. From the
        example below, the value False will be used for the column active::

            active = Column(Boolean, default=False)

        :param obj: A instance or a list of SQLAlchemy model instances.
        :param include: tuple, list or set.
        :param exclude: tuple, list or set.
        """

        if isinstance(obj, (list, InstrumentedList)):
            try:
                return [item.export_data(include, exclude) for item in obj]
            except AttributeError as e:
                # If the method exist, the exception comes inside of it.
                if hasattr(obj[0], "export_data"):
                    # So re-raise the exception.
                    raise e

                return [self(item, include, exclude) for item in obj]

        try:
            persisted = inspect(obj).persistent
        except NoInspectionAvailable as e:
            raise ValueError("Pass a valid SQLAlchemy mapped class instance")

        columns = obj.__mapper__.columns
        exclude = tuple(exclude) + self.exclude
        data = {}

        for c in columns:
            name = c.name

            if (not include or name in include) and name not in exclude:
                column_value = getattr(obj, name)

                data[name] = (
                    column_value
                    if persisted
                    else _get_column_default(c)
                    if column_value is None
                    else column_value
                )

        if persisted is True:
            unloaded_relationships = inspect(obj).unloaded
            relationship_keys = [
                relationship.key
                for relationship in obj.__class__.__mapper__.relationships
            ]

            for key in relationship_keys:
                if key not in unloaded_relationships and key not in exclude:
                    rproperty = getattr(obj, key)
                    has_export_data = hasattr(rproperty, "export_data")
                    data[key] = None

                    if has_export_data:
                        data[key] = rproperty.export_data()
                    elif rproperty:
                        data[key] = self(rproperty)

        return data


export_from_sqla_object = ExportData(exclude=("org_id",))


def generate_schema(model_class, include=(), exclude=()):
    """Inspect a SQLAlchemy Model Class and return a validation schema
    to be used with the Cerberus library. The schema is generated mapping some
    Cerberus rules with SQLAlchemy model class column types and constraints, as
    follow:

    ================ =======================================================
    Cerberus Rule    Generated value
    ================ =======================================================
    type             Based on the SQLAlchemy column class used (String,
                     Integer, etc).
    readonly         **True** if the column is primary key.
    required         **True** if the constraints ``Column.nullable`` is set to
                     **False** and ``Column.default`` and
                     ``Column.server_default`` are set to **None**.
    default          Not included in the output. This is handled by
                     SQLAlchemy or by the databse engine.
    ================ =======================================================
    """

    schema = {}

    mapper = inspect(model_class)

    for column in mapper.c:

        name = column.name

        if len(include) > 0 and name not in include:
            continue

        if name in exclude:
            continue

        prop = {}

        python_type = column.type.python_type

        prop["type"] = schema_type_conversions.get(python_type)

        if prop["type"] is None:
            raise LookupError("Unable to determine the column type")

        if python_type == str and column.type.length is not None:
            prop["maxlength"] = column.type.length

        if column.primary_key is True:
            prop["readonly"] = True

        if (
            column.default is None
            and column.server_default is None
            and column.nullable is False
            and column.primary_key is False
        ):
            prop["required"] = True

        schema[name] = prop

    return schema


def json(func):
    """Decorator for view functions to return a JSON response.

    When the incoming request is a POST request, it validates the content_type
    and payload before calling the view function. Next, the returned value of
    the view function is transformed into a JSON response.

    The view function can return the response payload, status code and headers
    in the next ways:

    1.  A single object. Can be any JSON serializable object, a Flask Response
        object, or a SQLAlchemy model:

        .. code-block:: python

            return {}

            # custom Response
            return make_response(...)

            # SQLAlchemy model instance
            return Mode.query.filter_by(prop=prop).first()

            return []

            return "string response"

    2.  A tuple in the form **(payload, status, headers)**, or **(payload, headers)**
        the response payload can be any python built-in type, or a SQLAlchemy based
        model object.:

        .. code-block:: python

            # payload, status

            return {}, 201

            return [], 201

            return '...', 400

            # payload, status, headers

            return {}, 201, {'X-Header': 'content'}

            # payload, headers

            return {}, {'X-Header': 'content'}
    """

    @wraps(func)
    def wrapper(*args, **kwargs):

        if request.method == "POST":
            if not request.is_json:
                abort(415, "application/json mimetype expected")

            if request.get_json(silent=True) is None:
                abort(400, "The request payload has an invalid JSON object")

        ro = func(*args, **kwargs)  # returned object

        if type(ro) != tuple:
            if isinstance(ro, (int, bool, str, list, dict, list)):
                return jsonify(ro)

            if isinstance(ro, Response):
                return ro

            ro = (ro,)

        payload, status, headers = ro + (None,) * (3 - len(ro))

        if type(status) != int:
            headers, status = status, None

        if is_sqla_obj(payload):
            try:
                payload = payload.export_data()
            except AttributeError as e:
                # If the method exist, the exception comes inside of it.
                if hasattr(payload, "export_data"):
                    # So re-raise the exception.
                    raise e

                payload = export_from_sqla_object(payload)

        response_object = jsonify(payload)

        if status:
            response_object.status_code = status

        if headers:
            response_object.headers.extend(headers)

        return response_object

    return wrapper


class Validator(_Validator):
    def __init__(self, schema, model_class=None, **kwargs):
        super(Validator, self).__init__(schema, **kwargs)
        self.model_class = model_class

    def validate(self, document, model=None, **kwargs):

        self.model = model

        return super(Validator, self).validate(document, **kwargs)

    def _validate_unique(self, is_unique, field, value):
        """Performs a query to the database to check value is already present
        in a given column.

        The rule's arguments are validated against this schema:
        {'type': 'boolean'}
        """

        if is_unique:
            if not self.model_class:
                raise RuntimeError(
                    "The rule `unique` needs a SQLAlchemy declarative class"
                    " to perform queries to check if the value being validated"
                    " is unique. Provide a class in Validator constructor."
                )

            filters = {field: value}
            model = self.model_class.query.filter_by(**filters).first()

            if model and (not self.update or model is not self.model):
                self._error(field, f"Must be unique, but '{value}' already exist")


def get_key_path(key, _map):

    for map_key, value in _map.items():
        path = []

        if map_key == key:
            return [map_key]

        if type(value) == dict:
            _path = get_key_path(key, value)
            path = ([map_key] + path + _path) if _path else []

        if len(path) > 0:
            return path

    return None



import jwt
from contextlib import contextmanager
from datetime import datetime, timedelta

from sqlalchemy import Column, Integer, String, DateTime, Boolean
from sqlalchemy import ForeignKey, func
from sqlalchemy.orm import relationship

from saraki.auth import _request_ctx_stack, User, Org
from saraki.model import BaseModel, Model, database


class DummyBaseModel(BaseModel):
    __tablename__ = "dummy_base_model"

    id = Column(Integer, primary_key=True)


class DummyModel(Model):
    __tablename__ = "dummy_model"

    id = Column(Integer, primary_key=True)


class Person(Model):

    __tablename__ = "person"

    id = Column(Integer, primary_key=True)

    firstname = Column(String, nullable=False)

    lastname = Column(String, nullable=False)

    age = Column(Integer, nullable=False)

    def export_data(self, include=("id", "firstname"), exclude=()):
        return super(Person, self).export_data(include, exclude)


class Product(BaseModel):

    __tablename__ = "product"

    id = Column(Integer, primary_key=True)

    name = Column(String(120), nullable=False)

    color = Column(String, default="white")

    price = Column(Integer, default=0)

    created_at = Column(DateTime, nullable=False, default=func.now())

    updated_at = Column(DateTime, nullable=False, server_default=func.now())

    enabled = Column(Boolean, default=False)


class Order(BaseModel):

    __tablename__ = "order"

    id = Column(Integer, primary_key=True)

    customer_id = Column(Integer, ForeignKey("person.id"), nullable=False)

    lines = relationship("OrderLine")

    customer = relationship("Person", uselist=False)


class OrderLine(Model):

    __tablename__ = "order_line"

    order_id = Column(Integer, ForeignKey("order.id"), nullable=False, primary_key=True)

    product_id = Column(
        Integer, ForeignKey("product.id"), nullable=False, primary_key=True
    )

    unit_price = Column(Integer, nullable=False)

    quantity = Column(Integer, default=1, nullable=False)

    product = relationship("Product", uselist=False)

    def export_data(self, include=(), exclude=()):
        include = tuple(include) + ("product_id", "unit_price", "quantity")
        return super(OrderLine, self).export_data(include, exclude)


class Cartoon(Model):

    __tablename__ = "cartoon"

    id = Column(Integer, primary_key=True)

    name = Column(String(80), unique=True, nullable=False)

    nickname = Column(String(80), unique=True)


class Todo(Model):

    __tablename__ = "todo"

    id = Column(Integer, primary_key=True)

    org_id = Column(Integer, ForeignKey("org.id"), nullable=False)

    task = Column(String(200), nullable=False)


def login(username, orgname=None, scope=None):
    iat = datetime.utcnow()
    exp = iat + timedelta(seconds=6000)
    payload = {"iss": "acme.local", "sub": username, "iat": iat, "exp": exp}

    if orgname:
        payload.update({"aud": orgname, "scp": {"org": ["manage"]}})

    if scope:
        payload.update({"scp": scope})

    token = jwt.encode(payload, "secret").decode()

    return f"JWT {token}"


@contextmanager
def auth_ctx(username, orgname=None):

    _request_ctx_stack.top.current_user = User(id=1, username=username)

    if orgname:
        _request_ctx_stack.top.current_org = Org(id=1, orgname=orgname)

    yield


def reset_secuence(table, column_name="id", schema_name="public"):

    table_name = f"{schema_name}.{table.__tablename__}"

    sql = f"SELECT pg_get_serial_sequence('{table_name}', '{column_name}');"
    secuence_name = database.engine.execute(sql).fetchone()[0]

    if secuence_name is not None:
        sql = f"ALTER SEQUENCE {secuence_name} RESTART WITH 1;"
        database.engine.execute(sql)



import pytest
from unittest.mock import patch, MagicMock
from json import loads, dumps
from assertions import list_is
from cerberus import Validator
from sqlalchemy.orm import joinedload
from saraki.model import User, Org, Membership, Resource, Action, Ability
from saraki.utility import generate_schema
from saraki.api import ORG_SCHEMA
from saraki.testing import assert_allowed_methods

from common import login


pytestmark = pytest.mark.usefixtures("data")


@patch("saraki.api.Validator")
def test_add_org_endpoint_data_validation(MockValidator, client):

    v = MagicMock()
    MockValidator.return_value = v
    v.validate.return_value = False
    v.errors = {}
    access_token = login("coyote")

    rv = client.post(
        "/users/coyote/orgs",
        data=dumps({"prop": "value"}),
        content_type="application/json",
        headers={"Authorization": access_token},
    )

    assert rv.status_code == 400

    MockValidator.assert_called_once_with(ORG_SCHEMA, Org)
    v.validate.assert_called_once_with({"prop": "value"})


@pytest.mark.usefixtures("data_org")
@pytest.mark.parametrize(
    "req_payload, status_code",
    [
        ({}, 400),
        ({"orgname": "acme", "name": "Acme Corporation"}, 400),
        ({"orgname": "choco", "name": "The Chocolate Factory"}, 201),
    ],
)
def test_add_org_endpoint(req_payload, status_code, client):

    rv = client.post(
        "/users/coyote/orgs",
        data=dumps(req_payload),
        content_type="application/json",
        headers={"Authorization": login("coyote")},
    )

    assert rv.status_code == status_code

    if rv.status_code == 201:
        org = (
            Org.query.options(joinedload(Org.created_by))
            .filter_by(orgname=req_payload["orgname"])
            .one()
        )

        member = Membership.query.filter_by(
            org_id=org.id, user_id=org.created_by.id
        ).one()

        assert org.orgname == "choco"
        assert org.name == "The Chocolate Factory"
        assert org.created_by.username == "coyote"
        assert member.is_owner is True


@pytest.mark.usefixtures("data_org")
@pytest.mark.parametrize(
    "username, expected_lst", [("coyote", [{"orgname": "acme"}]), ("YoseSam", [])]
)
def test_list_user_orgs_endpoint(client, username, expected_lst):

    token = login(username)
    url = f"/users/{username}/orgs"
    rv = client.get(url, headers={"Authorization": token})

    assert rv.status_code == 200

    returned_lst = loads(rv.data)

    assert len(expected_lst) == len(returned_lst)
    assert list_is(expected_lst) <= returned_lst


user_response_schema = generate_schema(
    User, exclude=["id", "password", "canonical_username"]
)


member_response_schema = {
    "user": {"type": "dict", "required": True, "schema": user_response_schema},
    "is_owner": {"type": "boolean", "required": True},
    "enabled": {"type": "boolean", "required": True},
}


@pytest.mark.usefixtures("data_member")
def test_list_members(client):
    token = login("coyote", "acme", scope={"org": ["read"]})

    rv = client.get("/orgs/acme/members", headers={"Authorization": token})

    assert rv.status_code == 200

    data = loads(rv.data)
    assert len(data) is 3

    v = Validator(member_response_schema)

    assert v.validate(data[0]), v.errors


@pytest.mark.usefixtures("data_org")
@pytest.mark.parametrize(
    "username, status, error",
    [
        ("unknown", 400, "User unknown does not exist"),
        ("coyote", 400, "coyote is already a member of acme"),
    ],
)
def test_add_member_with_invalid_user(client, username, status, error):
    data = {"username": username}
    token = login("coyote", "acme")

    rv = client.post(
        "/orgs/acme/members",
        data=dumps(data),
        content_type="application/json",
        headers={"Authorization": token},
    )

    body = loads(rv.data)

    assert rv.status_code == status
    assert error in body["error"]["username"]


@pytest.mark.usefixtures("data_org")
def test_add_member(client):
    data = {"username": "RoadRunner"}
    token = login("coyote", "acme")

    rv = client.post(
        "/orgs/acme/members",
        data=dumps(data),
        content_type="application/json",
        headers={"Authorization": token},
    )

    assert rv.status_code == 201

    v = Validator(member_response_schema)
    data = loads(rv.data)
    assert v.validate(data), v.errors


def test_list_plans(client):

    rv = client.get("/plans")
    rv.status_code == 200

    data = loads(rv.data)
    assert len(data) == 3


@pytest.mark.usefixtures("data")
def test_get_plan(client):

    rv = client.get("/plans/1")
    rv.status_code == 200

    data = loads(rv.data)
    assert data["id"] == 1


@pytest.mark.usefixtures("data")
def test_add_plan(client):
    token = login("coyote", scope={"app": ["write"]})
    data = {"name": "Enterprise", "amount_of_members": 500, "price": 1000}

    rv = client.post(
        "/plans",
        data=dumps(data),
        content_type="application/json",
        headers={"Authorization": token},
    )

    assert rv.status_code == 201, rv.data

    data = loads(rv.data)


def test_delete_plan(client):
    token = login("coyote", scope={"app": ["delete"]})
    url = "/plans/1"

    rv = client.delete(url, headers={"Authorization": token})

    assert rv.status_code == 200
    assert client.get(url).status_code == 404


class TestResource:
    def test_allowed_methods(self, app):
        assert_allowed_methods("/resources", ["GET"], app)
        assert_allowed_methods("/resources/1", ["GET"], app)

    def test_list_resource(self, client):
        token = login("coyote", scope={"app": "read"})
        rv = client.get("/resources", headers={"Authorization": token})

        assert rv.status_code == 200

        data = rv.get_json()

        assert len(data) > 0

    def test_get_resource(self, client):

        _id = Resource.query.first().id

        token = login("coyote", scope={"app": "read"})
        rv = client.get(f"/resources/{_id}", headers={"Authorization": token})

        assert rv.status_code == 200

        data = rv.get_json()
        assert data["id"] == _id


class TestAction:
    def test_allowed_methods(self, app):
        assert_allowed_methods("/actions", ["GET"], app)
        assert_allowed_methods("/actions/1", ["GET"], app)

    def test_list_resource(self, client):
        token = login("coyote", scope={"app": "read"})
        rv = client.get("/actions", headers={"Authorization": token})

        assert rv.status_code == 200

        data = rv.get_json()
        assert len(data) > 0

    def test_get_resource(self, client):
        _id = Action.query.first().id

        token = login("coyote", scope={"app": "read"})
        rv = client.get(f"/actions/{_id}", headers={"Authorization": token})

        assert rv.status_code == 200

        data = rv.get_json()
        assert data["id"] == _id


class TestAbility:
    def test_allowed_methods(self, app):
        assert_allowed_methods("/abilities", ["GET"], app)
        assert_allowed_methods("/abilities/1,1", ["GET"], app)

    def test_list_abilities(self, client):
        token = login("coyote", scope={"app": ["read"]})
        rv = client.get("/abilities", headers={"Authorization": token})

        assert rv.status_code == 200

        data = rv.get_json()
        assert len(data) > 0

    def test_get_ability(self, client):
        model = Ability.query.first()
        _id = f"{model.action_id},{model.resource_id}"

        token = login("coyote", scope={"app": "read"})
        rv = client.get(f"/abilities/{_id}", headers={"Authorization": token})

        assert rv.status_code == 200

        data = rv.get_json()
        assert data["action_id"] == model.action_id
        assert data["resource_id"] == model.resource_id


@pytest.mark.usefixtures("data_role")
class TestRole:
    def test_allowed_methods(self, app):
        assert_allowed_methods("/orgs/acme/roles", ["GET", "POST"], app)
        assert_allowed_methods("/orgs/acme/roles/1", ["GET", "PATCH", "DELETE"], app)

    @pytest.mark.parametrize("orgname, expected", [("acme", 3), ("rrinc", 4)])
    def test_list_roles(self, client, orgname, expected):
        token = login("coyote", orgname, scope={"org": ["manage"]})
        rv = client.get(f"/orgs/{orgname}/roles", headers={"Authorization": token})
        data = rv.get_json()

        assert rv.status_code == 200
        assert len(data) == expected

    @pytest.mark.parametrize(
        "orgname, id, expected",
        [("acme", 5, 404), ("acme", 1, 200)],
        ids=["from another organization", "own role"],
    )
    def test_get_role(self, client, orgname, id, expected):
        token = login("coyote", orgname, scope={"org": ["manage"]})
        rv = client.get(f"/orgs/{orgname}/roles/{id}", headers={"Authorization": token})

        assert rv.status_code == expected

    def test_add_role_with_invalid_payload(self, client):
        token = login("coyote", "acme", scope={"org": ["manage"]})
        data = dumps({"name": "New Role", "description": "...", "org_id": 4})
        rv = client.post(
            f"/orgs/acme/roles",
            headers={"Authorization": token},
            data=data,
            content_type="application/json",
        )
        data = rv.get_json()

        assert rv.status_code == 400
        assert "unknown field" in data["error"]["org_id"]

    def test_add_role_without_abilities(self, client):
        token = login("coyote", "acme", scope={"org": ["manage"]})
        data = dumps({"name": "New Role", "description": "..."})
        rv = client.post(
            f"/orgs/acme/roles",
            headers={"Authorization": token},
            data=data,
            content_type="application/json",
        )
        data = rv.get_json()

        assert rv.status_code == 201
        assert "org_id" not in data
        assert "id" in data
        assert data["name"] == "New Role"
        assert data["description"] == "..."

    def test_add_role_with_abilities(self, client):
        token = login("coyote", "acme", scope={"org": ["manage"]})
        data = dumps(
            {
                "name": "New Role",
                "description": "...",
                "abilities": [
                    {"action_id": 1, "resource_id": 1},
                    {"action_id": 2, "resource_id": 1},
                ],
            }
        )
        rv = client.post(
            f"/orgs/acme/roles",
            headers={"Authorization": token},
            data=data,
            content_type="application/json",
        )
        data = rv.get_json()

        assert rv.status_code == 201, rv.data
        assert "org_id" not in data
        assert "id" in data
        assert data["name"] == "New Role"
        assert data["description"] == "..."
        assert len(data["abilities"]) == 2

    def test_update_role_from_another_organization(self, client):
        token = login("coyote", "acme", scope={"org": ["manage"]})
        req_payload = {"name": "Updated Role 5", "description": "New description"}
        rv = client.patch(
            f"/orgs/acme/roles/5",
            headers={"Authorization": token},
            data=dumps(req_payload),
            content_type="application/json",
        )

        assert rv.status_code == 404

    def test_update_role_with_invalid_payload(self, client):
        token = login("coyote", "acme", scope={"org": ["manage"]})
        req_payload = {"org_id": 2}
        rv = client.patch(
            f"/orgs/acme/roles/1",
            headers={"Authorization": token},
            data=dumps(req_payload),
            content_type="application/json",
        )
        data = rv.get_json()

        assert rv.status_code == 400
        assert "unknown field" in data["error"]["org_id"]

    def test_update_role_without_abilities(self, client):
        token = login("coyote", "acme", scope={"org": ["manage"]})
        req_payload = {"name": "Updated Role 1", "description": "New description"}
        rv = client.patch(
            f"/orgs/acme/roles/1",
            headers={"Authorization": token},
            data=dumps(req_payload),
            content_type="application/json",
        )
        data = rv.get_json()

        assert rv.status_code == 200
        assert data["name"] == req_payload["name"]
        assert data["description"] == req_payload["description"]

    def test_update_role_with_abilities(self, client):

        token = login("coyote", "acme", scope={"org": ["manage"]})
        data = dumps(
            {
                "name": "New Role",
                "description": "...",
                "abilities": [
                    {"action_id": 1, "resource_id": 1},
                    {"action_id": 1, "resource_id": 2},
                ],
            }
        )
        rv = client.post(
            "/orgs/acme/roles",
            headers={"Authorization": token},
            data=data,
            content_type="application/json",
        )
        data = rv.get_json()

        assert rv.status_code == 201

        _id = data["id"]

        data = dumps(
            {
                "name": "Updated Role",
                "abilities": [
                    {"action_id": 1, "resource_id": 1},
                    {"action_id": 1, "resource_id": 3},
                ],
            }
        )
        rv = client.patch(
            f"/orgs/acme/roles/{_id}",
            headers={"Authorization": token},
            data=data,
            content_type="application/json",
        )
        data = rv.get_json()

        assert rv.status_code == 200

    @pytest.mark.parametrize(
        "id, expected",
        [(1, 200), (5, 404)],
        ids=["own role", "from another organization"],
    )
    def test_delete_role(self, client, id, expected):
        token = login("coyote", "acme", scope={"org": ["manage"]})

        rv = client.delete(f"/orgs/acme/roles/{id}", headers={"Authorization": token})
        assert rv.status_code == expected

        rv = client.get(f"/orgs/acme/roles/{id}", headers={"Authorization": token})
        assert rv.status_code == 404


@pytest.mark.usefixtures("data_member_role")
class TestMemberRole:
    def test_list_member_roles(self, client):
        token = login("coyote", "acme", scope={"org": ["manage"]})
        rv = client.get(
            "/orgs/acme/members/yosesam/roles", headers={"Authorization": token}
        )
        data = rv.get_json()

        assert rv.status_code == 200
        assert len(data) > 0

        role = data[0]
        data = rv.get_json()
        assert "name" in role
        assert "abilities" in role

    @pytest.mark.parametrize(
        "orgname, username, id, expected",
        [
            ("acme", "yosesam", 2, 200),
            ("acme", "yosesam", 1, 404),
            ("rrinc", "yosesam", 4, 404),
            ("rrinc", "yosesam", 6, 200),
        ],
    )
    def test_get_member_roles(self, client, orgname, username, id, expected):
        token = login(username, orgname, scope={"org": ["manage"]})
        rv = client.get(
            f"/orgs/{orgname}/members/{username}/roles/{id}",
            headers={"Authorization": token},
        )

        assert rv.status_code == expected

    def test_get_member_roles_repsonse_payload(self, client):
        token = login("yosesam", "acme", scope={"org": ["manage"]})
        rv = client.get(
            f"/orgs/acme/members/yosesam/roles/2", headers={"Authorization": token}
        )

        assert rv.status_code == 200

    def test_add_role_to_member(self, client):
        token = login("coyote", "acme", scope={"org": ["manage"]})
        rv = client.post(
            "/orgs/acme/members/yosesam/roles",
            headers={"Authorization": token},
            content_type="application/json",
            data=dumps({"roles": [1, 3]}),
        )

        assert rv.status_code == 201

    def test_delete_role_from_member(self, client):
        token = login("coyote", "acme", scope={"org": ["manage"]})
        rv = client.delete(
            "/orgs/acme/members/yosesam/roles/2", headers={"Authorization": token}
        )
        assert rv.status_code == 200

        rv = client.get(
            "/orgs/acme/members/yosesam/roles/2", headers={"Authorization": token}
        )
        assert rv.status_code == 404



import jwt
import pytest
from random import randint
from typing import NamedTuple
from json import dumps, loads
from werkzeug.exceptions import BadRequest, NotFound
from flask import Flask
from calendar import timegm
from datetime import datetime, timedelta
from unittest.mock import patch, call, Mock
from saraki.exc import (
    NotFoundCredentialError,
    InvalidPasswordError,
    InvalidUserError,
    JWTError,
    AuthorizationError,
    TokenNotFoundError,
    InvalidMemberError,
    InvalidOrgError,
    ProgrammingError,
)
from saraki.auth import (
    _verify_username,
    _verify_orgname,
    _verify_member,
    _authenticate_with_password,
    _authenticate_with_token,
    _get_request_jwt,
    _generate_jwt_payload,
    _encode_jwt,
    _decode_jwt,
    _authenticate,
    _is_authorized,
    _validate_request,
    require_auth,
    Auth,
    current_user,
    current_org,
)
from saraki.model import User, Org

parametrize = pytest.mark.parametrize


def getpayload(scp=None, sub="coyote", aud=None):
    iat = datetime.utcnow()
    exp = iat + timedelta(seconds=6000)
    payload = {"iss": "acme.local", "sub": sub, "iat": iat, "exp": exp}

    if aud:
        payload["aud"] = aud

    if scp:
        payload["scp"] = scp

    return payload


@pytest.fixture
def _payload():
    return getpayload()


class AppUserMock(NamedTuple):
    username: str


class AppOrgMock(NamedTuple):
    orgname: str


def _get_user(username):
    return AppUserMock(username=username)


def _get_org(orgname):
    return AppOrgMock(orgname=orgname)


@pytest.mark.usefixtures("data")
class Test_verify_username(object):
    def test_different_case_variations(self, ctx):
        assert _verify_username("coyote").username == "coyote"
        assert _verify_username("Coyote").username == "coyote"
        assert _verify_username("COYOTE").username == "coyote"

    def test_unregistered_username(self, ctx):
        error_msg = 'Username "unknown" is not registered'

        with pytest.raises(InvalidUserError, match=error_msg):
            _verify_username("unknown")


@pytest.mark.usefixtures("data", "data_org")
class Test_verify_orgname:
    def test_registered_organization(self, ctx):
        assert _verify_orgname("acme").orgname == "acme"

    def test_unregistered_organization(self, ctx):

        error_msg = 'Orgname "unknown" is not registered'

        with pytest.raises(InvalidOrgError, match=error_msg):
            _verify_orgname("unknown")


@pytest.mark.usefixtures("data", "data_org")
class Test_verify_member:
    def test_valid_member(self, ctx):

        user = _verify_username("coyote")
        org = _verify_orgname("acme")

        member = _verify_member(user, org)
        assert member.user is user
        assert member.org is org

    def test_invali_member(self, ctx):

        user = _verify_username("coyote")
        org = _verify_orgname("rrinc")

        error_msg = "coyote is not a member of rrinc"

        with pytest.raises(InvalidMemberError, match=error_msg):
            _verify_member(user, org)


class Test_get_request_jwt(object):
    def test_with_more_than_one_space(self, request_ctx):
        headers = {"Authorization": "Bearer this is a token with spaces"}

        with request_ctx("/", headers=headers):
            with pytest.raises(JWTError, match="The token contains spaces"):
                _get_request_jwt()

    def test_with_wrong_prefix(self, request_ctx):
        headers = {"Authorization": "WRONG a.nice.token"}

        with request_ctx("/", headers=headers):
            with pytest.raises(JWTError, match="Unsupported authorization type"):
                _get_request_jwt()

    def test_without_space_after_prefix(self, request_ctx):
        headers = {"Authorization": "Bearera.nice.token"}

        with request_ctx("/", headers=headers):
            with pytest.raises(JWTError, match="Missing or malformed token"):
                _get_request_jwt()

    def test_with_empty_authorization_http_header(self, request_ctx):
        with request_ctx("/", headers={"Authorization": ""}):
            with pytest.raises(JWTError, match="Missing or malformed token"):
                _get_request_jwt()

    def test_without_authorization_http_header(self, request_ctx):
        with request_ctx("/"):
            token = _get_request_jwt()

        assert token is None

    def test_with_valid_token(self, app):
        prefix = app.config["JWT_AUTH_HEADER_PREFIX"]
        token = "a.nice.token"
        headers = {"Authorization": f"{prefix} {token}"}

        with app.test_request_context("/", headers=headers):
            _token = _get_request_jwt()

        assert token == _token


class Test_generate_jwt_payload(object):
    @pytest.mark.usefixtures("ctx")
    def test_default_included_claims(self):
        user = _get_user(username="somebody")

        payload = _generate_jwt_payload(user)

        assert len(payload) is 3
        assert "iat" in payload
        assert "exp" in payload
        assert "sub" in payload

    def test_iss_claim_when_no_JWT_ISSUER_or_SERVER_NAME_are_set(self, app):
        app.config["SERVER_NAME"] = None
        app.config["JWT_ISSUER"] = None
        app.config["JWT_REQUIRED_CLAIMS"] = ["iss"]

        error_msg = (
            "The token payload could not be generated. The claim iss is "
            "required, but neither JWT_ISSUER nor SERVER_NAME are provided"
        )

        user = _get_user(username="somebody")

        with app.app_context():
            with pytest.raises(RuntimeError, match=error_msg):
                _generate_jwt_payload(user)

    def test_iss_claim_inclusion_only_when_required(self, app):
        app.config["JWT_REQUIRED_CLAIMS"] = []
        app.config["SERVER_NAME"] = "server.name"
        app.config["JWT_ISSUER"] = "acme.issuer"

        user = _get_user(username="somebody")

        with app.app_context():
            payload = _generate_jwt_payload(user)

        assert "iss" not in payload

    def test_iss_claim(self, app):
        app.config["JWT_REQUIRED_CLAIMS"] = ["iss"]
        app.config["SERVER_NAME"] = "server.name"
        app.config["JWT_ISSUER"] = "acme.issuer"

        user = _get_user(username="somebody")

        with app.app_context():
            payload = _generate_jwt_payload(user)

        assert payload["iss"] == "acme.issuer"

    def test_iss_claim_fallback_to_server_name(self, app):
        app.config["JWT_REQUIRED_CLAIMS"] = ["iss"]
        app.config["SERVER_NAME"] = "server.name"
        app.config["JWT_ISSUER"] = None

        user = _get_user(username="somebody")

        with app.app_context():
            payload = _generate_jwt_payload(user)

        assert payload["iss"] == "server.name"

    @pytest.mark.usefixtures("ctx")
    def test_sub_claim(self):
        user = _get_user(username="coyote")
        payload = _generate_jwt_payload(user)

        assert payload["sub"] == "coyote"

    @pytest.mark.usefixtures("data", "data_org")
    def test_aud_claim(self, app):
        user = User.query.filter_by(username="coyote").one()
        org = Org.query.filter_by(orgname="acme").one()

        with app.app_context():
            payload = _generate_jwt_payload(user, org)

        assert payload["aud"] == "acme"

    @pytest.mark.usefixtures("ctx")
    @patch("saraki.auth.datetime")
    def test_iat_claim(self, mock_datetime):
        _datetime = datetime(2018, 5, 13, 17, 52, 44, 524300)
        mock_datetime.utcnow.return_value = _datetime

        user = _get_user(username="somebody")
        payload = _generate_jwt_payload(user)

        assert payload["iat"] == _datetime
        mock_datetime.utcnow.assert_called_once()

    @patch("saraki.auth.datetime")
    def test_exp_claim(self, mock_datetime, app):
        _datetime = datetime(2018, 5, 13, 17, 52, 44, 524300)
        app.config["JWT_EXPIRATION_DELTA"] = timedelta(seconds=400)
        mock_datetime.utcnow.return_value = _datetime

        user = _get_user(username="somebody")

        with app.app_context():
            payload = _generate_jwt_payload(user)

        assert payload["exp"] == _datetime + timedelta(seconds=400)

    @pytest.mark.usefixtures("data", "data_org")
    def test_scp_claim_for_organization_owners(self):

        user = User.query.filter_by(username="coyote").one()
        org = Org.query.filter_by(orgname="acme").one()

        payload = _generate_jwt_payload(user, org)

        assert "org" in payload["scp"]
        assert payload["scp"]["org"] == ["manage"]


class Test_encode_jwt(object):
    @pytest.mark.usefixtures("ctx")
    def test_with_valid_payload(self):
        iat = datetime.utcnow()
        exp = iat + timedelta(seconds=300)

        token = _encode_jwt(
            {
                "iss": "acme.local",
                "sub": "coyote",
                "iat": iat,
                "exp": exp,
                "custom": "custom claim value",
            }
        )

        _payload = jwt.decode(token, verify=False)

        assert _payload["iss"] == "acme.local"
        assert _payload["sub"] == "coyote"
        assert _payload["iat"] == timegm(iat.utctimetuple())
        assert _payload["exp"] == timegm(exp.utctimetuple())
        assert _payload["custom"] == "custom claim value"

    def test_with_missing_iss_claim(self, app):
        app.config["JWT_REQUIRED_CLAIMS"] = ["iss"]
        iat = datetime.utcnow()
        exp = iat + timedelta(seconds=300)

        error = "Payload is missing required claims: iss"

        with app.app_context():
            with pytest.raises(ValueError, match=error):
                _encode_jwt({"sub": "coyote", "iat": iat, "exp": exp})

    @pytest.mark.usefixtures("ctx")
    def test_with_various_missing_required_claims(self):
        with pytest.raises(ValueError) as error:
            _encode_jwt({})

        e = str(error)

        assert "Payload is missing required claims:" in e
        assert "iss" in e
        assert "sub" in e
        assert "iat" in e
        assert "exp" in e

    def test_when_a_secret_is_not_provided(self, app):

        app.config["SECRET_KEY"] = None

        error = "SECRET_KEY is not set. Can not generate Token"

        with app.app_context():
            with pytest.raises(RuntimeError, match=error):
                _encode_jwt({})


class Test_decode_jwt(object):
    def test_passing_invalid_data_types(self):

        with pytest.raises(ValueError, match="is not a valid JWT string"):
            _decode_jwt(None)

        with pytest.raises(ValueError, match="is not a valid JWT string"):
            _decode_jwt(1)

        with pytest.raises(ValueError, match="is not a valid JWT string"):
            _decode_jwt(True)

    def test_token_with_wrong_iss_claim(self, app, _payload):
        app.config["JWT_ISSUER"] = "acme.local"
        app.config["JWT_LEEWAY"] = timedelta(seconds=10)
        app.config["JWT_REQUIRED_CLAIMS"] = ["iss"]

        _payload["iss"] = "malicious.issuer"

        token = jwt.encode(_payload, app.config["SECRET_KEY"])

        with app.app_context():
            with pytest.raises(JWTError, match="Invalid issuer"):
                _decode_jwt(token)

    def test_token_without_iss_claim_when_is_required(self, app, _payload):
        app.config["JWT_REQUIRED_CLAIMS"] = ["iss"]
        app.config["JWT_ISSUER"] = "acme.local"
        app.config["JWT_LEEWAY"] = timedelta(seconds=10)
        del _payload["iss"]

        token = jwt.encode(_payload, app.config["SECRET_KEY"])
        error_msg = 'Token is missing the "iss" claim'

        with app.app_context():
            with pytest.raises(JWTError, match=error_msg):
                _decode_jwt(token)

    def test_token_without_iat_claim(self, app, _payload):

        app.config["JWT_LEEWAY"] = timedelta(seconds=10)
        del _payload["iat"]

        token = jwt.encode(_payload, app.config["SECRET_KEY"])
        error_msg = 'Token is missing the "iat" claim'

        with app.app_context():
            with pytest.raises(JWTError, match=error_msg):
                _decode_jwt(token)

    def test_token_without_exp_claim(self, app, _payload):

        app.config["JWT_LEEWAY"] = timedelta(seconds=10)
        del _payload["exp"]

        token = jwt.encode(_payload, app.config["SECRET_KEY"])
        error_msg = 'Token is missing the "exp" claim'

        with app.app_context():
            with pytest.raises(JWTError, match=error_msg):
                _decode_jwt(token)

    def test_with_expired_token(self, app):
        app.config["JWT_LEEWAY"] = timedelta(seconds=10)

        iat = datetime(1900, 5, 13, 17, 52, 44, 524300)
        exp = iat + timedelta(seconds=400)
        payload = {"iat": iat, "exp": exp}

        token = jwt.encode(payload, app.config["SECRET_KEY"])

        with app.app_context():
            with pytest.raises(JWTError, match="Token has expired"):
                _decode_jwt(token)

    def test_with_malformed_token(self, app, _payload):

        token = jwt.encode(_payload, app.config["SECRET_KEY"]) + b"'"

        with app.app_context():
            with pytest.raises(JWTError, match="Invalid or malformed token"):
                _decode_jwt(token)

    def test_with_valid_token_without_iss(self, app, _payload):
        app.config["JWT_LEEWAY"] = timedelta(seconds=10)

        token = jwt.encode(_payload, app.config["SECRET_KEY"])

        with app.app_context():
            decoded_payload = _decode_jwt(token)

        assert decoded_payload["sub"] == "coyote"

    def test_with_valid_token(self, app, _payload):
        app.config["SERVER_NAME"] = "acme.local"
        app.config["JWT_LEEWAY"] = timedelta(seconds=10)

        token = jwt.encode(_payload, app.config["SECRET_KEY"])

        with app.app_context():
            decoded_payload = _decode_jwt(token)

        assert decoded_payload["iss"] == "acme.local"
        assert decoded_payload["sub"] == "coyote"

    def test_with_valid_token_with_aud_claim(self, app):
        app.config["SERVER_NAME"] = "acme.local"
        app.config["JWT_LEEWAY"] = timedelta(seconds=10)

        payload = getpayload(aud="acme")
        token = jwt.encode(payload, app.config["SECRET_KEY"])

        with app.app_context():
            decoded_payload = _decode_jwt(token)

        assert decoded_payload["iss"] == "acme.local"
        assert decoded_payload["sub"] == "coyote"
        assert decoded_payload["aud"] == "acme"


class Test_is_authorized(object):
    def test_sub_claim_verification(self, app, _payload):
        @app.route("/<sub:username>/private-info")
        def private_info(username):
            return "Private information"

        _payload["sub"] = "coyote"

        with app.test_request_context("/elmer/private-info"):
            assert _is_authorized(_payload) is False

        _payload["sub"] = "elmer"

        with app.test_request_context("/elmer/private-info"):
            assert _is_authorized(_payload) is True

    @parametrize(
        "payload, expected",
        [
            (getpayload(), False),
            (getpayload(aud="unknown"), False),
            (getpayload(aud="acme"), True),
        ],
        ids=["Missing aud claim", "aud value mismatch", "aud value matches"],
    )
    def test_aud_claim_verification(self, app, payload, expected):
        @app.route("/<aud:orgname>/private-info")
        def private_info(orgname):
            return "Private information"

        with app.test_request_context("/acme/private-info"):
            assert _is_authorized(payload) is expected

    def test_routes_without_required_claim(self, app, _payload):

        # route without any view_args
        @app.route("/private-info")
        def list_private():
            return "Private information"

        # route with view_args
        @app.route("/private-info/<int:id>")
        def get_private(id):
            return "Private information"

        with app.test_request_context("/private-info"):
            assert _is_authorized(_payload) is True

        with app.test_request_context("/private-info/1"):
            assert _is_authorized(_payload) is True

    @parametrize(
        "scope, required, expected",
        [
            (None, (None, None), True),
            (None, ("movie", "delete"), False),
            ({"film": ["read"]}, (None, None), True),
            ({"film": ["read"]}, ("movie", "delete"), False),
            ({"film": ["manage"]}, ("movie", "delete"), False),
            ({"movie": ["read"]}, ("movie", "delete"), False),
            ({"movie": ["delete"]}, ("movie", "delete"), True),
            ({"movie": ["manage"]}, ("movie", "delete"), True),
        ],
        ids=[
            "without scope | resource not required",
            "without scope | resource is required",
            "with scope | resource not required",
            "missing required resource",
            "manage action but missing required resource",
            "resource match but action does not",
            "resource and action match",
            "resource match and action is manage",
        ],
    )
    def test_scope_validation(self, app, scope, required, expected):
        @app.route("/")
        def private_info():
            return ""

        payload = getpayload(scp=scope)

        resource = required[0]
        action = required[1]

        with app.test_request_context("/"):
            assert _is_authorized(payload, resource, action) is expected

    @parametrize(
        "scope, required, expected",
        [
            (None, ("catalog", "read"), False),
            ({"catalog": ["read"]}, ("catalog", "read"), True),
            ({"product": ["read"]}, ("catalog", "read"), True),
            ({"catalog": ["read"]}, ("catalog", "write"), False),
            ({"product": ["read"]}, ("catalog", "write"), False),
            ({"product": ["write"]}, ("product", "write"), True),
        ],
        ids=[
            "without scope | resource is required",
            "resource and action match",
            "resource is parent",
            "resource match but action does not",
            "resource is parent but action does not match",
            "scope and required resource are parents",
        ],
    )
    def test_scope_validation_with_nested_resource(
        self, app, scope, required, expected
    ):

        app.auth._resources = {"puchase": None, "product": {"catalog": None}}

        @app.route("/")
        def private_info():
            return ""

        payload = getpayload(scp=scope)

        resource = required[0]
        action = required[1]

        with app.test_request_context("/"):
            assert _is_authorized(payload, resource, action) is expected

    @parametrize(
        "action, method, expect",
        [
            ("custom", "CUSTOM", False),
            ("manage", "CUSTOM", False),
            ("follow", "GET", False),
            ("read", "DELETE", False),
            ("write", "GET", False),
            ("read", "GET", True),
            ("write", "POST", True),
            ("write", "PUT", True),
            ("write", "PATCH", True),
            ("delete", "DELETE", True),
            ("manage", "GET", True),
            ("manage", "POST", True),
            ("manage", "PUT", True),
            ("manage", "PATCH", True),
            ("manage", "DELETE", True),
        ],
    )
    def test_http_method_to_action_mapping(self, app, action, method, expect):
        @app.route("/", methods=[method])
        def private_info():
            return ""

        payload = getpayload(scp={"cartoon": [action]})

        with app.test_request_context("/", method=method):
            assert _is_authorized(payload, "cartoon") is expect


class Test_validate_request(object):
    def test_request_without_access_token(self, request_ctx):
        with request_ctx("/"):
            with pytest.raises(TokenNotFoundError):
                _validate_request()

    @patch("saraki.auth._verify_username")
    @patch("saraki.auth._is_authorized")
    @patch("saraki.auth._decode_jwt")
    @patch("saraki.auth._get_request_jwt")
    def test_request_with_access_token(
        self,
        mocked_get_request_jwt,
        mocked_jwt_decode_handler,
        mocked_is_authorized,
        mocked_verify_username,
        request_ctx,
    ):

        mocked_get_request_jwt.return_value = "a.nice.token"
        mocked_jwt_decode_handler.return_value = {"sub": "coyote"}

        with request_ctx("/"):
            _validate_request("stock", "read")

        mocked_get_request_jwt.assert_called_once()
        mocked_jwt_decode_handler.assert_called_once_with("a.nice.token")
        mocked_is_authorized.assert_called_once_with({"sub": "coyote"}, "stock", "read")
        mocked_verify_username.assert_called_once_with("coyote")

    @patch("saraki.auth._is_authorized")
    def test_with_unauthorized_token(self, mocked_is_authorized, app):

        token = jwt.encode(getpayload(), app.config["SECRET_KEY"]).decode()
        headers = {"Authorization": f"JWT {token}"}

        mocked_is_authorized.return_value = False

        with app.test_request_context("/", headers=headers):
            with pytest.raises(AuthorizationError):
                _validate_request()

    @pytest.mark.usefixtures("data")
    def test_with_unknown_username_in_sub_claim(self, app, _payload):

        path = f"/{randint(100, 10000)}"

        @app.route(path)
        def index():
            pass

        _payload["sub"] = "unknown"

        token = jwt.encode(_payload, app.config["SECRET_KEY"]).decode()
        headers = {"Authorization": f"JWT {token}"}

        with app.test_request_context(path, headers=headers):
            with pytest.raises(AuthorizationError):
                _validate_request()

    @pytest.mark.usefixtures("data")
    def test_with_unknown_orgname_in_aud_claim(self, app):
        path = f"/{randint(100, 10000)}"

        @app.route(path)
        def index():
            pass

        payload = getpayload(aud="unknown")

        token = jwt.encode(payload, app.config["SECRET_KEY"]).decode()
        headers = {"Authorization": f"JWT {token}"}

        with app.test_request_context(path, headers=headers):
            with pytest.raises(AuthorizationError):
                _validate_request()

    @pytest.mark.usefixtures("data")
    def test_current_user(self, app):
        @app.route("/")
        def index():
            pass

        payload = getpayload(sub="coyote")

        token = jwt.encode(payload, app.config["SECRET_KEY"]).decode()
        headers = {"Authorization": f"JWT {token}"}

        with app.test_request_context("/", headers=headers):
            _validate_request()
            assert current_user.username == "coyote"

        assert current_user._get_current_object() is None

    @pytest.mark.usefixtures("data", "data_org")
    def test_current_org(self, app):
        @app.route("/")
        def index():
            pass

        payload = getpayload(sub="coyote", aud="acme")

        token = jwt.encode(payload, app.config["SECRET_KEY"]).decode()
        headers = {"Authorization": f"JWT {token}"}

        with app.test_request_context("/", headers=headers):
            _validate_request()
            assert current_org.orgname == "acme"

        assert current_org._get_current_object() is None


@pytest.mark.usefixtures("data")
class Test_authenticate_with_password(object):
    @pytest.mark.usefixtures("ctx")
    def test_with_unknown_username(self):
        error = 'Username "extraneous" is not registered'

        with pytest.raises(InvalidUserError, match=error):
            _authenticate_with_password("extraneous", "12345")

    @pytest.mark.usefixtures("ctx")
    def test_with_wrong_password(self, ctx):

        with pytest.raises(InvalidPasswordError, match="Invalid password"):
            _authenticate_with_password("coyote", "wrongpassword")

    @pytest.mark.usefixtures("ctx")
    def test_with_valid_username_and_password(self):
        identity = _authenticate_with_password("coyote", "12345")

        assert identity.username == "coyote"


@pytest.mark.usefixtures("data")
class Test_authenticate_with_token(object):
    def test_token_with_unregistered_username(self, app, _payload):

        _payload["sub"] = "unknown"

        token = jwt.encode(
            _payload, app.config["SECRET_KEY"], algorithm=app.config["JWT_ALGORITHM"]
        )

        error = 'Username "unknown" is not registered'

        with app.app_context():
            with pytest.raises(InvalidUserError, match=error):
                _authenticate_with_token(token)

    def test_token_with_registered_username(self, app, _payload):

        token = jwt.encode(
            _payload, app.config["SECRET_KEY"], algorithm=app.config["JWT_ALGORITHM"]
        )

        with app.app_context():
            user = _authenticate_with_token(token)

        assert user.username == "coyote"


class Test_authenticate(object):
    def test_request_without_credentials_or_token(self, request_ctx):

        with request_ctx("/"):
            with pytest.raises(NotFoundCredentialError):
                _authenticate()

    def test_request_without_username(self, request_ctx):

        body = dumps({"password": "12345"})

        with request_ctx("/", data=body, content_type="application/json"):
            with pytest.raises(BadRequest):
                _authenticate()

    def test_request_without_password(self, request_ctx):

        body = dumps({"username": "coyote"})

        with request_ctx("/", data=body, content_type="application/json"):
            with pytest.raises(BadRequest):
                _authenticate()

    @pytest.mark.usefixtures("data")
    def test_request_passing_unregistered_orgname(self, request_ctx):

        body = dumps({"username": "Y0seSam", "password": "secret"})

        with request_ctx("/", data=body, content_type="application/json"):
            with pytest.raises(NotFound):
                _authenticate("unknown")

    @pytest.mark.usefixtures("data", "data_org")
    def test_request_passing_registered_orgname(self, request_ctx):

        body = dumps({"username": "coyote", "password": "12345"})

        with request_ctx("/", data=body, content_type="application/json"):
            rv = _authenticate("acme")

        assert rv.status_code == 200
        assert "access_token" in loads(rv.data)


@pytest.mark.usefixtures("data")
class TestAuthenticationRequest(object):
    """Integration test of ``/auth`` and ``/auth/<orgname>`` endpoints."""

    @parametrize(
        "req_payload",
        [
            (None),
            ({}),
            ({"password": "12345"}),
            ({"username": "coyote"}),
            ({"username": "coyote", "password": "wrongpassword"}),
            ({"username": "unknown", "password": "123456"}),
        ],
        ids=[
            "Empty request payload",
            "Empty dictionary",
            "Missing username",
            "Missing password",
            "Wrong password",
            "Unregistered username",
        ],
    )
    def test_invalid_authentication_request(self, client, req_payload):

        rv = client.post(
            "/auth", data=dumps(req_payload), content_type="application/json"
        )

        assert rv.status_code == 400

    def test_request_with_valid_credentials(self, client):

        body = {"username": "coyote", "password": "12345"}

        rv = client.post("/auth", data=dumps(body), content_type="application/json")

        assert rv.status_code == 200
        token = loads(rv.data)["access_token"]
        payload = jwt.decode(token, verify=False)

        assert payload["sub"] == "coyote"

    def test_request_with_valid_token(self, app, _payload):
        token = jwt.encode(
            _payload, app.config["SECRET_KEY"], algorithm=app.config["JWT_ALGORITHM"]
        ).decode()

        rv = app.test_client().post(
            "/auth",
            headers={"Authorization": f"JWT {token}"},
            content_type="application/json",
        )

        assert rv.status_code == 200
        token = loads(rv.data)["access_token"]
        payload = jwt.decode(token, verify=False)

        assert payload["sub"] == "coyote"

    @pytest.mark.usefixtures("data", "data_org")
    @parametrize(
        "req_payload",
        [
            (None),
            ({}),
            ({"password": "12345"}),
            ({"username": "coyote"}),
            ({"username": "coyote", "password": "wrongpassword"}),
            ({"username": "unknown", "password": "123456"}),
            ({"username": "R0adRunner", "password": "password"}),
        ],
        ids=[
            "Empty request payload",
            "Empty dictionary",
            "Missing username",
            "Missing password",
            "Wrong password",
            "Unregistered username",
            "Not a member",
        ],
    )
    def test_invalid_org_authentication_request(self, client, req_payload):

        rv = client.post(
            "/auth/acme", data=dumps(req_payload), content_type="application/json"
        )

        assert rv.status_code == 400

    def test_unregistered_org_authentication_request(self, client):

        orgname = "unknown"
        request_payload = {"username": "Coyote", "password": "123456"}

        rv = client.post(
            f"/auth/{orgname}",
            data=dumps(request_payload),
            content_type="application/json",
        )

        assert rv.status_code == 404

    @pytest.mark.usefixtures("data", "data_org")
    @parametrize(
        "payload, expected",
        [(getpayload(sub="R0adRunner"), 400), (getpayload(sub="coyote"), 200)],
        ids=["Not a member", "Owner"],
    )
    def test_org_auth_request_using_valid_tokens(self, app, payload, expected):

        token = jwt.encode(
            payload, app.config["SECRET_KEY"], algorithm=app.config["JWT_ALGORITHM"]
        ).decode()

        rv = app.test_client().post(
            "/auth/acme",
            headers={"Authorization": f"JWT {token}"},
            content_type="application/json",
        )

        assert rv.status_code == expected

        if rv.status_code == 200:
            token = loads(rv.data)["access_token"]
            payload = jwt.decode(token, verify=False)

            assert payload["sub"] == "coyote"
            assert payload["aud"] == "acme"


class TestRequireAuth(object):
    def test_pass_action_without_resource(self):
        error_msg = "You passed an action 'read' without a resource"

        with pytest.raises(ProgrammingError, match=error_msg):
            require_auth(action="read")

    @parametrize(
        "resource, action, parent_resource",
        [("stock", None, None), ("stock", "read", None), ("stock", "read", "market")],
    )
    def test_auth_metadata(self, resource, action, parent_resource):
        @require_auth(resource, action, parent_resource)
        def endpoint():
            return "private content"

        assert hasattr(endpoint, "_auth_metadata")
        assert endpoint._auth_metadata == {
            "resource": resource,
            "action": action,
            "parent_resource": parent_resource,
        }

    @patch("saraki.auth._validate_request")
    @parametrize("resource, action", [(None, None), ("stock", None), ("stock", "read")])
    def test_call_to_validate_request_internal_function(
        self, mocked_validate_request, resource, action
    ):
        @require_auth(resource, action)
        def private():
            return "private content"

        private()

        mocked_validate_request.assert_called_once_with(resource, action)


class TestAuth(object):
    def test_instantiation(self):
        auth = Auth()
        assert not hasattr(auth, "app")

        with patch.object(Auth, "init_app") as mocked_init_app:
            app = Flask(__name__)
            auth = Auth(app)
            mocked_init_app.assert_called_once_with(app)

    def test_init_app(self):
        app = Flask(__name__)
        auth = Auth()
        auth.init_app(app)

        adapter = app.url_map.bind("")

        assert adapter.match("/auth", method="POST")
        assert auth.app is app

    def test_add_resource(self):
        auth = Auth()
        auth._add_resource("lonely")

        assert "lonely" in auth.resources

        auth = Auth()
        auth._add_resource("purchase", "commerce")

        assert "commerce" in auth.resources
        assert "purchase" in auth.resources["commerce"]

        auth = Auth()
        auth._add_resource("product")
        auth._add_resource("catalog")
        auth._add_resource("product")

        assert len(auth.resources) == 2
        assert "catalog" in auth.resources
        assert "product" in auth.resources

        auth = Auth()
        auth._add_resource("sale", "commerce")
        auth._add_resource("purchase", "commerce")

        assert auth.resources == {"commerce": {"sale": None, "purchase": None}}

    @parametrize("value", (None, 1, True, [], ()))
    def test_add_resouce_with_invalid_value(self, value):
        auth = Auth()

        with pytest.raises(TypeError):
            auth._add_resource(value)

    @parametrize(
        "actions, length",
        [
            (("read", "update", "read", "update", "create"), 4),
            (("read", "create", "update", "delete"), 5),
        ],
    )
    def test_add_action(self, actions, length):
        auth = Auth()

        for action_name in actions:
            auth._add_action(action_name)

        assert len(auth.actions) == length

    @parametrize("value", (None, 1, True, [], ()))
    def test_add_action_with_invalid_value(self, value):
        auth = Auth()

        with pytest.raises(TypeError):
            auth._add_action(value)

    def test_collect_metadata(self):
        app = Flask(__name__)
        auth = Auth(app)

        @app.route("/sales")
        @require_auth("sale", "read", "biz")
        def list_sale():
            pass

        @app.route("/sales", methods=["POST"])
        @require_auth("sale", "create", "biz")
        def add_sale():
            pass

        @app.route("/sales/1")
        @require_auth()
        def get_one_sale():
            pass

        auth._add_resource = Mock()
        auth._add_action = Mock()

        auth._collect_metadata()

        assert auth._add_resource.call_count == 2
        auth._add_resource.assert_has_calls([call("sale", "biz"), call("sale", "biz")])

        assert auth._add_resource.call_count == 2
        auth._add_action.assert_has_calls([call("read"), call("create")])


@pytest.mark.usefixtures("data")
class TestEndpoint(object):
    @parametrize(
        "payload, expected",
        [
            (None, 401),
            (getpayload(sub="unknown"), 401),
            (getpayload(sub="coyote"), 200),
        ],
    )
    def test_route_without_sub_variable_rule(self, app, payload, expected):

        client = app.test_client()

        @app.route("/movies")
        @require_auth()
        def movies():
            return "movies"

        headers = {}

        if payload:
            token = jwt.encode(
                payload, app.config["SECRET_KEY"], algorithm=app.config["JWT_ALGORITHM"]
            )
            headers["Authorization"] = f"JWT {token.decode()}"

        assert client.get("/movies", headers=headers).status_code == expected

    @parametrize(
        "payload, expected",
        [
            (None, 401),
            (getpayload(sub="unknown"), 401),
            (getpayload(sub="R0adRunner"), 401),
            (getpayload(sub="coyote"), 200),
        ],
    )
    def test_route_with_sub_variable_rule(self, app, payload, expected):

        client = app.test_client()

        @app.route("/<sub:username>/my-movies")
        @require_auth()
        def my_movies(username):
            return "Those are your movies"

        headers = {}

        if payload:
            token = jwt.encode(
                payload, app.config["SECRET_KEY"], algorithm=app.config["JWT_ALGORITHM"]
            )

            headers["Authorization"] = f"JWT {token.decode()}"

        rv = client.get("/coyote/my-movies", headers=headers)
        assert rv.status_code == expected

    @pytest.mark.usefixtures("data", "data_org")
    @parametrize(
        "payload, expected",
        [
            (None, 401),
            (getpayload(sub="coyote", aud="unknown"), 401),
            (getpayload(sub="coyote", aud="acme"), 200),
        ],
    )
    def test_route_with_aud_variable_rule(self, app, payload, expected):
        @app.route("/<aud:orgname>/stocks")
        @require_auth()
        def stocks(orgname):
            return "All the stocks"

        headers = {}

        if payload:
            token = jwt.encode(
                payload, app.config["SECRET_KEY"], algorithm=app.config["JWT_ALGORITHM"]
            )

            headers["Authorization"] = f"JWT {token.decode()}"

        client = app.test_client()
        rv = client.get("/acme/stocks", headers=headers)
        assert rv.status_code == expected

    @parametrize(
        "payload, expected",
        [
            (None, 401),
            (getpayload(), 200),
            (getpayload({"film": ["manage"]}), 200),
            (getpayload({"movie": ["write"]}), 200),
        ],
        ids=[
            "without token",
            "token without scope",
            "token with manage action",
            "token with different action",
        ],
    )
    def test_without_required_resource(self, app, payload, expected):

        client = app.test_client()

        @app.route("/movies")
        @require_auth()
        def movies():
            return "movies"

        headers = {}

        if payload:
            token = jwt.encode(
                payload, app.config["SECRET_KEY"], algorithm=app.config["JWT_ALGORITHM"]
            )
            headers["Authorization"] = f"JWT {token.decode()}"

        assert client.get("/movies", headers=headers).status_code == expected

    @parametrize(
        "payload, expected",
        [
            (None, 401),
            (getpayload(), 401),
            (getpayload({"film": ["manage"]}), 401),
            (getpayload({"movie": ["read"]}), 401),
            (getpayload({"movie": ["write"]}), 200),
        ],
        ids=[
            "without token",
            "token: without scope",
            "token: missing required resource",
            "token: action mismatch",
            "token: with required resource/action",
        ],
    )
    def test_with_required_resource(self, app, payload, expected):
        @app.route("/movies", methods=["POST"])
        @require_auth("movie")
        def movies():
            return "movies"

        headers = {}

        if payload:
            token = jwt.encode(
                payload, app.config["SECRET_KEY"], algorithm=app.config["JWT_ALGORITHM"]
            )
            headers["Authorization"] = f"JWT {token.decode()}"

        client = app.test_client()
        assert client.post("/movies", headers=headers).status_code == expected

    @parametrize(
        "payload, expected",
        [
            (None, 401),
            (getpayload(), 401),
            (getpayload({"film": ["manage"]}), 401),
            (getpayload({"film": ["write"]}), 401),
            (getpayload({"movie": ["change"]}), 200),
        ],
    )
    def test_with_explicit_required_action(self, app, payload, expected):

        client = app.test_client()

        @app.route("/movies", methods=["PUT"])
        @require_auth("movie", "change")
        def movies():
            return "movies"

        headers = {}

        if payload:
            token = jwt.encode(
                payload, app.config["SECRET_KEY"], algorithm=app.config["JWT_ALGORITHM"]
            )
            headers["Authorization"] = f"JWT {token.decode()}"

        assert client.put("/movies", headers=headers).status_code == expected


def test_register_persist_actions_handler():
    auth = Auth()

    @auth.persist_actions
    def custom_handler(actions):
        pass

    assert auth._persist_actions_func == custom_handler


def test_register_persist_resources_handler():
    auth = Auth()

    @auth.persist_resources
    def custom_handler(resources):
        pass

    assert auth._persist_resources_func == custom_handler



import pytest
from common import Product
from saraki.utility import generate_schema


@pytest.fixture(scope="module")
def _schema():
    return generate_schema(Product)


def test_included_properties(_schema):

    assert len(_schema) == 7
    assert "id" in _schema
    assert "name" in _schema
    assert "color" in _schema
    assert "price" in _schema
    assert "created_at" in _schema
    assert "updated_at" in _schema
    assert "enabled" in _schema


def test_primary_key_column(_schema):
    assert _schema["id"]["readonly"] is True


def test_integer_data_type(_schema):
    assert _schema["id"]["type"] == "integer"
    assert _schema["price"]["type"] == "integer"


def test_string_data_type_with_max_length(_schema):
    assert _schema["name"]["type"] == "string"
    assert _schema["name"]["maxlength"] == 120


def test_string_data_type_without_max_length(_schema):
    assert _schema["color"]["type"] == "string"
    assert "maxlength" not in _schema["color"]


def test_datetime_data_type(_schema):
    assert _schema["created_at"]["type"] == "string"
    assert _schema["updated_at"]["type"] == "string"


def test_boolean_data_type(_schema):
    assert _schema["enabled"]["type"] == "boolean"


def test_non_nullable_columns(_schema):
    assert _schema["name"]["required"] is True


def test_non_nullable_column_with_default_value(_schema):
    assert "required" not in _schema["created_at"]


def test_non_nullable_columns_with_server_default_value(_schema):
    assert "required" not in _schema["updated_at"]


def test_include_argument():
    schema = generate_schema(Product, include=["id", "enabled"])

    assert len(schema) == 2
    assert "id" in schema
    assert "enabled" in schema


def test_exclude_argument():
    schema = generate_schema(Product, exclude=["id", "enabled"])

    assert len(schema) == 5
    assert "id" not in schema
    assert "enabled" not in schema



# -*- coding: utf-8 -*-

"""Top-level package for Saraki."""

__author__ = """José María Domínguez Moreno"""
__email__ = "miso.0b11@gmail.com"
__version__ = "0.1.0a0"


from saraki.app import Saraki, Blueprint  # noqa: F401
from saraki.auth import require_auth  # noqa: F401



import pytest
from saraki.model import (
    User,
    Org,
    Resource,
    Action,
    _persist_actions,
    _persist_resources,
    get_member_privileges,
)


class Test_persist_actions:
    @pytest.mark.usefixtures("ctx", "savepoint")
    def test_add_new_actions(self):

        # Amount of actions registered in source code may change overtime.
        # So base this test on the current amount of resources.
        current_length = len(Action.query.all())

        data = ["follow", "print"]
        _persist_actions(data)

        actions = Action.query.all()
        assert len(actions) == current_length + 2

    @pytest.mark.usefixtures("ctx", "savepoint")
    def test_repeated_actions(self):
        data = {"follow", "print"}
        _persist_actions(data)

        # Amount of actions registered in source code may change overtime.
        # So base this test on the current amount of resources.

        current_length = len(Action.query.all())

        data |= {"follow", "erase", "spy"}
        _persist_actions(data)

        actions = Action.query.all()
        assert len(actions) == current_length + 2


class Test_persist_resources:
    @pytest.mark.usefixtures("ctx", "savepoint")
    def test_add_new_resources(self):

        # Amount of resources registered in source code may change overtime.
        # So base this test on the current amount of resources.
        current_length = len(Resource.query.all())

        data = {"purchase": None, "product": {"catalog": None}}

        _persist_resources(data)
        resources = {r.name: r for r in Resource.query.all()}
        assert len(resources) == 3 + current_length

        assert resources["purchase"].parent is None
        assert resources["product"].parent is None
        assert resources["catalog"].parent is resources["product"]

    @pytest.mark.usefixtures("ctx", "savepoint")
    def test_repeated_resources(self):

        # Amount of resources registered in source code may change overtime.
        # So base this test on the current amount of resources.
        current_length = len(Resource.query.all())

        data = {"purchase": None, "product": {"catalog": None}}

        _persist_resources(data)

        resources = Resource.query.all()
        assert len(resources) == 3 + current_length

        # repeat purchase, product and catalog
        data = {
            "sales": None,
            "purchase": None,
            "product": {"catalog": None, "descontinued": None},
        }

        _persist_resources(data)

        resources = {r.name: r for r in Resource.query.all()}
        assert len(resources) == 5 + current_length

        assert resources["sales"].parent is None
        assert resources["purchase"].parent is None
        assert resources["product"].parent is None
        assert resources["catalog"].parent is resources["product"]
        assert resources["descontinued"].parent is resources["product"]


@pytest.mark.usefixtures("data", "data_member_role")
class Test_get_member_privileges:
    def test_when_member_is_owner(self, ctx):
        user = User.query.filter_by(canonical_username="coyote").one()
        org = Org.query.filter_by(orgname="acme").one()

        privileges = get_member_privileges(org, user)

        assert privileges == {"org": ["manage"]}

    def test_member_without_privileges(self, ctx):

        user = User.query.filter_by(canonical_username="yosesam").one()
        org = Org.query.filter_by(orgname="acme").one()

        privileges = get_member_privileges(org, user)

        assert privileges == {}

    def test_member_with_privileges(self, ctx):

        user = User.query.filter_by(canonical_username="roadrunner").one()
        org = Org.query.filter_by(orgname="acme").one()

        privileges = get_member_privileges(org, user)

        assert privileges == {"org": ["read", "write"]}



import os
from datetime import timedelta


SECRET_KEY = os.getenv("SECRET_KEY")

SERVER_NAME = os.getenv("SERVER_NAME")

SQLALCHEMY_DATABASE_URI = os.getenv("DATABASE_URI")

SQLALCHEMY_TRACK_MODIFICATIONS = False

JWT_LEEWAY = timedelta(seconds=10)

JWT_ALGORITHM = "HS256"

JWT_EXPIRATION_DELTA = timedelta(seconds=300)

JWT_AUTH_HEADER_PREFIX = "JWT"

JWT_ISSUER = SERVER_NAME

JWT_REQUIRED_CLAIMS = ["exp", "iat", "sub"]



from werkzeug.exceptions import NotFound
from werkzeug.routing import RequestRedirect, MethodNotAllowed
from sqlalchemy import event


def get_view_function(url, method="GET", app=None):
    adapter = app.url_map.bind("")

    try:
        match = adapter.match(url, method=method)
    except RequestRedirect as e:
        # recursively match redirects
        return get_view_function(e.new_url, method)
    except (MethodNotAllowed, NotFound):
        # no match
        return None

    try:
        # return the view function and arguments
        return app.view_functions[match[0]], match[1]
    except KeyError:
        # no view is associated with the endpoint
        return None


def assert_allowed_methods(path, methods, app):
    """Ensures that a URL only allows a set of HTTP methods.

    It not only checks that the HTTP methods passed in the ``methods`` parameter
    are allowed, but also that the URL does not allow methods not included in
    the list.

    If the path provided does not exist a :class:`werkzeug.exceptions.NotFound`
    exception is raised.

    :param path: The string URL to test.
    :param methods: List of HTTP methods expected to be allowed.
    :param app: A Flask application instance.
    """

    adapter = app.url_map.bind("")

    # Get the list of allowed methods
    current_methods = adapter.allowed_methods(path)

    # If the list is empty is because no route matches the path.
    if not current_methods:
        raise NotFound(
            f"{path}. Make sure that an endpoint is implemented that handles it."
        )

    current_methods.remove("OPTIONS")

    # Check if HEAD is present because just endpoints that implements GET
    # explicily implement the HEAD method.
    if "HEAD" in current_methods:
        current_methods.remove("HEAD")

    for current_method in current_methods:
        assert (
            current_method in methods
        ), "The path `{}` should not allow the method {}".format(
            path, current_method
        )

    for expected_method in methods:
        assert (
            expected_method in current_methods
        ), "The path `{}` does not implement the method {}".format(
            path, expected_method
        )


class Savepoint(object):
    """It helps to interact with the database in isolation and then return to
    a previously known state, discarding changes such as the insertion, deletion
    and update of records.::

        with app.app_context():
            sv = Savepoint(database)
            sv.start()

            model = Cartoon()
            database.session.add(model)
            database.session.commit()

            sv.end()  # Rollback all changes

    """

    def __init__(self, database):
        self.database = database
        self.connection = None

    def started(self):
        return self.connection and not self.connection.closed

    def start(self):
        """Starts a root transaction and uses PostgreSQL SAVEPOINTs to keep
        database modification around without really committing the changes to
        the database.
        """

        if self.started():
            raise RuntimeError("There is already an ongoing transaction.")

        connection = self.database.engine.connect()

        # Begin a non-ORM transaction
        root_transaction = connection.begin()

        options = dict(bind=self.database.engine)
        session = self.database.create_scoped_session(options=options)

        self.database.session = session

        # Each time a SAVEPOINT ends, create a new one.
        @event.listens_for(session, "after_transaction_end")
        def restart_savepoint(session, transaction):
            is_savepoint = transaction.nested

            if is_savepoint and not transaction._parent.nested:
                session.expire_all()
                session.begin_nested()

        # Create the first SAVEPOINT
        session.begin_nested()

        self.root_transaction = root_transaction
        self.connection = connection
        self.session = session

    def end(self):
        # Use remove() istead of close() because this is an scoped session.
        self.session.remove()
        self.root_transaction.rollback()
        self.connection.close()



from saraki.model import (
    database,
    Plan,
    User,
    Org,
    Membership,
    Action,
    Role,
    RoleAbility,
    Resource,
    MemberRole,
)
from common import Person, Product, Order, OrderLine, Cartoon, reset_secuence


def insert_persons():
    lst = [
        {"id": 1, "firstname": "Nikola", "lastname": "Tesla", "age": 24},
        {"id": 2, "firstname": "Albert", "lastname": "Einstein", "age": 21},
        {"id": 3, "firstname": "Isaac", "lastname": "Newton", "age": 12},
    ]

    database.session.add_all([Person(**item) for item in lst])


def insert_products():
    lst = [
        {"name": "Explosive Tennis Balls", "color": "white", "price": 9},
        {"name": "Binocular", "color": "black", "price": 99},
        {"name": "Acme anvils", "color": "black", "price": 999},
    ]

    database.session.add_all([Product(**item) for item in lst])


def insert_orders():

    # Orders
    order_lst = [{"id": 1, "customer_id": 1}, {"id": 2, "customer_id": 2}]
    database.session.add_all([Order(**item) for item in order_lst])

    # Order lines
    order_line_lst = [
        {"order_id": 1, "product_id": 1, "quantity": 3, "unit_price": 14},
        {"order_id": 1, "product_id": 2, "quantity": 4, "unit_price": 142},
        {"order_id": 1, "product_id": 3, "quantity": 7, "unit_price": 73},
        {"order_id": 2, "product_id": 1, "quantity": 2, "unit_price": 21},
    ]
    database.session.add_all([OrderLine(**item) for item in order_line_lst])


def insert_cartoons():
    lst = [
        {"name": "Bugs Bunny", "nickname": "bugs"},
        {"name": "Pepé Le Pew", "nickname": "pepe"},
        {"name": "Sylvester J. Pussycat Sr.", "nickname": "sylvester"},
    ]

    database.session.add_all([Cartoon(**item) for item in lst])


def insert_actions():
    reset_secuence(Action)

    lst = [
        {"name": "manage", "description": "manage description"},
        {"name": "read", "description": "read description"},
        {"name": "write", "description": "write description"},
        {"name": "delete", "description": "delete description"},
    ]

    database.session.add_all([Action(**item) for item in lst])
    database.session.commit()


def insert_resources():
    reset_secuence(Resource)

    lst = [
        {"name": "app", "description": "app description"},
        {"name": "org", "description": "org description"},
        {"name": "inventory", "description": "inventory description"},
    ]

    database.session.add_all([Resource(**item) for item in lst])
    database.session.commit()


def insert_plans():
    lst = [
        {"name": "Basic", "amount_of_members": 5, "price": 0},
        {"name": "RoadRunner", "amount_of_members": 10, "price": 20},
        {"name": "YoseSam", "amount_of_members": 50, "price": 100},
    ]

    database.session.add_all([Plan(**item) for item in lst])


def insert_users():
    lst = [
        {
            "username": "coyote",
            "email": "coyote@acme",
            "plain_text_password": "12345",
            "_password": (
                "$bcrypt-sha256$2b,"
                "12$lBIBMOF0u7gt6ruXNKalIO$i2MHLlTAbDNrw./RtN9jhDbdC/3dMXu"
            ),
        },
        {
            "username": "RoadRunner",
            "email": "RoadRunner@acme",
            "plain_text_password": "password",
            "_password": (
                "$bcrypt-sha256$2b,"
                "12$f9QIjKmLyaWb3gEp4KTORe$jrfpqWIANSyaEBnzQNz6YLcJusn9uH6"
            ),
        },
        {
            "username": "YoseSam",
            "email": "y0sesam@acme",
            "plain_text_password": "secret",
            "_password": (
                "$bcrypt-sha256$2b,"
                "12$3n9wunxWHfvZj0mg8v10Ru$lSo2YkVRepteVO62W8Q4L3Dn3xmoPna"
            ),
        },
    ]

    for item in lst:
        user = User(
            username=item["username"], email=item["email"], _password=item["_password"]
        )
        database.session.add(user)


def insert_orgs():
    lst = [
        {"username": "coyote", "orgname": "acme", "name": "Acme Corporation"},
        {"username": "RoadRunner", "orgname": "rrinc", "name": "RR Inc"},
    ]

    for data in lst:
        user = User.query.filter_by(username=data["username"]).one()

        org = Org(orgname=data["orgname"], name=data["name"], created_by=user)
        database.session.add(org)

        member = Membership(user=user, org=org, is_owner=True)

        database.session.add(member)

    database.session.commit()


def insert_members():
    acme = Org.query.filter_by(orgname="acme").one()
    rr = Org.query.filter_by(orgname="rrinc").one()

    yosesam = User.query.filter_by(username="YoseSam").one()
    runner = User.query.filter_by(username="RoadRunner").one()

    lst = [
        {"user_id": runner.id, "org_id": acme.id, "is_owner": False},
        {"user_id": yosesam.id, "org_id": acme.id, "is_owner": False},
        {"user_id": yosesam.id, "org_id": rr.id, "is_owner": False},
    ]

    database.session.add_all([Membership(**data) for data in lst])
    database.session.commit()


def insert_roles():
    reset_secuence(Role)

    acme = Org.query.filter_by(orgname="acme").one()
    rr = Org.query.filter_by(orgname="rrinc").one()

    lst = [
        # Acme
        {"name": "role 1", "description": "description 1", "org_id": acme.id},
        {"name": "role 2", "description": "description 2", "org_id": acme.id},
        {"name": "role 3", "description": "description 3", "org_id": acme.id},
        # R.R. Inc
        {"name": "role 4", "description": "description 4", "org_id": rr.id},
        {"name": "role 5", "description": "description 5", "org_id": rr.id},
        {"name": "role 6", "description": "description 6", "org_id": rr.id},
        {"name": "role 7", "description": "description 7", "org_id": rr.id},
    ]

    roles = [Role(**data) for data in lst]

    database.session.add_all(roles)

    role_abilities = [
        {"role": roles[0], "resource_id": 2, "action_id": 2},
        {"role": roles[0], "resource_id": 2, "action_id": 3},
    ]

    database.session.add_all([RoleAbility(**data) for data in role_abilities])

    database.session.commit()


def insert_member_roles():
    runner = User.query.filter_by(username="RoadRunner").one()
    yosesam = User.query.filter_by(username="YoseSam").one()

    acme = Org.query.filter_by(orgname="acme").one()
    rr = Org.query.filter_by(orgname="rrinc").one()

    lst = [
        {"role_id": 1, "user_id": runner.id, "org_id": acme.id},
        {"role_id": 2, "user_id": yosesam.id, "org_id": acme.id},
        {"role_id": 6, "user_id": yosesam.id, "org_id": rr.id},
    ]

    database.session.add_all([MemberRole(**data) for data in lst])
    database.session.commit()



import pytest
from assertions import list_is


dummy_lst = [{"name": "Elmer"}, {"name": "Sam"}]


params = (
    ("lst", "subset_lst", "expected"),
    [
        ([], [], True),
        ([{}, {}], [{}, {}, {}], True),
        (
            [{"id": 1, "name": "Jon", "pets": []}, {"id": 2, "name": "Sam"}],
            [{"id": 1, "pets": []}],
            True,
        ),
        ([{"id": 1}], [{"id": 1, "name": "Jon"}, {"id": 2, "name": "Sam"}], False),
        ([{"id": 1}], [{"id": 1, "name": "Elmer"}], False),
        (
            [{"id": 1, "name": "Elmer"}, {"id": 2, "name": "Sam"}],
            [{"id": 1, "name": "Elmer"}, {"id": 2, "name": "Sam"}],
            True,
        ),
        (dummy_lst, dummy_lst, True),
    ],
)


@pytest.mark.parametrize(*params)
def test_list_is_subset_of(lst, subset_lst, expected):

    if expected is True:
        assert list_is(subset_lst).subset_of(lst)
        assert list_is(subset_lst) <= lst
    else:
        assert not list_is(subset_lst).subset_of(lst)
        assert not list_is(subset_lst) <= lst


@pytest.mark.parametrize(*params)
def test_list_has_subset(lst, subset_lst, expected):

    if expected is True:
        assert list_is(lst).has_subset(subset_lst)
        assert list_is(lst) >= subset_lst
    else:
        assert not list_is(lst).has_subset(subset_lst)
        assert not list_is(lst) >= subset_lst



import pytest
from flask import Flask
from werkzeug.exceptions import NotFound
from saraki.model import database
from saraki.testing import Savepoint, assert_allowed_methods
from common import Cartoon


class Test_assert_allowed_methods:
    def test_non_existent_path(self):
        app = Flask(__name__)

        with pytest.raises(NotFound):
            assert_allowed_methods("/unknown", ["GET"], app)

    def test_path_allowing_unlisted_methods(self):
        app = Flask(__name__)

        @app.route("/candies", methods=["GET", "POST", "DELETE"])
        def candies():
            pass

        message = "The path `/candies` should not allow the method DELETE"

        with pytest.raises(AssertionError, match=message):
            assert_allowed_methods("/candies", ["GET", "POST"], app)

    def test_path_not_allowing_listed_methods(self):
        app = Flask(__name__)

        @app.route("/candies", methods=["GET"])
        def candies():
            pass

        message = "The path `/candies` does not implement the method POST"

        with pytest.raises(AssertionError, match=message):
            assert_allowed_methods("/candies", ["GET", "POST"], app)

    def test_path_implementing_only_listed_methods(self):
        app = Flask(__name__)

        @app.route("/candies", methods=["PATCH", "PUT"])
        def candies():
            pass

        assert_allowed_methods("/candies", ["PATCH", "PUT"], app)


@pytest.mark.usefixtures("data")
class TestSavepoint:
    def test_insert_rows(self, app):
        sv = Savepoint(database)

        with app.app_context():
            assert len(Cartoon.query.all()) == 3

        with app.app_context():
            sv.start()

            cartoon = Cartoon(name="aaa")
            database.session.add(cartoon)
            database.session.commit()

            assert len(Cartoon.query.all()) == 4

            sv.end()

        with app.app_context():
            assert len(Cartoon.query.all()) == 3

    def test_deleting_rows(self, app):
        sv = Savepoint(database)

        with app.app_context():
            assert len(Cartoon.query.all()) == 3

        with app.app_context():
            sv.start()

            cartoon = Cartoon.query.get(1)
            database.session.delete(cartoon)
            database.session.commit()

            assert len(Cartoon.query.all()) == 2

            sv.end()

        with app.app_context():
            assert len(Cartoon.query.all()) == 3

    def test_start_without_closing_current_transaction(self, app):
        sv = Savepoint(database)

        with app.app_context():
            message = "There is already an ongoing transaction."

            with pytest.raises(RuntimeError, match=message):
                sv.start()
                sv.start()



from passlib.hash import bcrypt_sha256
from sqlalchemy import (
    Column,
    ForeignKey,
    Integer,
    String,
    Boolean,
    Text,
    UniqueConstraint,
    ForeignKeyConstraint,
)
from sqlalchemy.orm import relationship, aliased
from sqlalchemy.ext.hybrid import hybrid_property
from flask_sqlalchemy import SQLAlchemy
from saraki.utility import import_into_sqla_object, export_from_sqla_object

database = SQLAlchemy()
BaseModel = database.Model


class ModelMixin:
    def import_data(self, data):
        return import_into_sqla_object(self, data)

    def export_data(self, include=(), exclude=()):
        return export_from_sqla_object(self, include, exclude)


class Model(BaseModel, ModelMixin):

    __abstract__ = True


class Plan(Model):
    """Plans available for your application."""

    __tablename__ = "plan"

    #: Primary key
    id = Column(Integer, primary_key=True)

    #: A name for the plan. For instance, Pro, Business, Personal, etc.
    name = Column(String(100), nullable=False, unique=True)

    #: The amount of members that an organization can have.
    amount_of_members = Column(Integer, nullable=False, default=1)

    #: Price of the plan.
    price = Column(Integer, nullable=False, unique=True)


class User(Model):
    """Application user accounts."""

    __tablename__ = "user"

    id = Column(Integer, primary_key=True)

    #: Email associated with the account. Must be unique.
    email = Column(String(128), unique=True, nullable=False)

    _username = Column("username", String(20), unique=True, nullable=False)

    @hybrid_property
    def username(self):
        """Username associated with the account. Must be unique."""
        return self._username

    @username.setter
    def username(self, value):
        self._username = value
        self._canonical_username = value.lower()

    _canonical_username = Column(
        "canonical_username", String(20), unique=True, nullable=False
    )

    @hybrid_property
    def canonical_username(self):
        """ Lowercase version of the username used for authentication.

        Don't set this column directly. This column is filled automatically
        when the :attr:`username` column is assigned with a value.
        """
        return self._canonical_username

    #: Don't set the value of this property directly, instead use the hybrid
    #: property ``password`` that will encrypt it for you.
    _password = Column("password", String(255), nullable=False)

    @hybrid_property
    def password(self):
        """ The password is hashed under the hood, so set this with the
        original/unhashed password directly.
        """
        return self._password

    @password.setter
    def password(self, value):
        self._password = bcrypt_sha256.hash(value)

    #: This property defines if the user account is activated or not. To use
    #: when the user verifies its account through an email for instance.
    active = Column(Boolean(), default=False, nullable=False)

    def verify_password(self, value):
        return bcrypt_sha256.verify(value, self.password)

    def import_data(self, data):
        super(User, self).import_data(data)

        # .import_data does not work with hybrid_property, so we set these
        # manually.
        if "username" in data:
            self.username = data["username"]

        if "password" in data:
            self.password = data["password"]

    def export_data(self, include=(), exclude=("id", "password", "canonical_username")):
        return super(User, self).export_data(include, exclude)

    def __str__(self):
        return f"<User {self.username}>"


class Org(Model):
    """Organization accounts registered and managed by the application.

    This table registers all organizations being managed by the application and
    owned by at least one user account registered in the :class:`Membership`
    table.
    """

    __tablename__ = "org"

    #: Primary Key.
    id = Column(Integer, primary_key=True)

    #: The organization account name.
    orgname = Column(String(20), unique=True, nullable=False)

    #: The name of the organization.
    name = Column(String(80), unique=True, nullable=False)

    #: The primary key of the user that created the organization account. But,
    #: this account not necessarily is the owner of the organization account,
    #: just the user that registered the organization. See the table
    #: :class:`Member` for more information.
    user_id = Column(Integer, ForeignKey("user.id"), nullable=False)

    #: Plan selected from the :class:`Plan` table.
    plan_id = Column(Integer, ForeignKey("plan.id"))

    #
    # - - - Relationships - - -
    #

    created_by = relationship("User", uselist=False)

    plan = relationship("Plan", uselist=False)


class Membership(Model):
    """Users accounts that are members of an Organization.

    Application users who belong to an organization are considered members,
    including the owner of the account. This table is a many to many
    relationship between the tables :class:`User` and :class:`Org`.
    """

    __tablename__ = "membership"

    #: The ID of a user account in the table :class:`User`.
    user_id = Column(Integer, ForeignKey("user.id"), primary_key=True)

    #: The ID of an organization account in the table :class:`Org`.
    org_id = Column(Integer, ForeignKey("org.id"), primary_key=True)

    #: If this is True, this member is the/an owner of this organization. One
    #: or more members can be owner at the same time.
    is_owner = Column(Boolean, default=False, server_default="FALSE", nullable=False)

    #: Enable or disable a member from an organization.
    enabled = Column(Boolean, default=False, nullable=False)

    #
    # -- Relationships --
    #

    org = relationship("Org", uselist=False)
    user = relationship("User", uselist=False)

    def export_data(self, include=(), exclude=("org_id", "user_id")):
        return super(Membership, self).export_data(include, exclude)


class Action(Model, ModelMixin):
    """Actions performed across the application like manage, create, read,
    update, delete, follow, etc.
    """

    __tablename__ = "action"

    id = Column(Integer, primary_key=True)

    name = Column(String(80), nullable=False, unique=True)

    description = Column(Text)

    def import_data(self, data):
        super(Action, self).import_data(data)


class Resource(Model, ModelMixin):
    """Application resources."""

    __tablename__ = "resource"

    #: Primary Key.
    id = Column(Integer, primary_key=True)

    #: The name of the resource.
    name = Column(String(80), nullable=False, unique=True)

    #: A useful description, please.
    description = Column(Text)

    #: Parent resource.
    parent_id = Column(Integer, ForeignKey("resource.id"))

    #
    # -- Relationships --
    #

    parent = relationship("Resource", uselist=False, remote_side=id)


def _persist_actions(actions):
    """Saves a list of actions in the database. Actions in the list that
    already exist in the database are ignored.

    Note that this doesn't commit the current session.

    :param actions: list, tuple or set of action names.
    """

    actions = set(actions)
    persisted = {action.name for action in Action.query.all()}
    new_actions = actions - persisted

    for name in new_actions:
        action = Action(name=name)
        database.session.add(action)


def _persist_resources(resources, parent=None):
    """Save resources in the database.

    Note that this doesn't commit the current session.
    """

    persisted = {r.name: r for r in Resource.query.all()}

    for key, value in resources.items():
        if key not in persisted:
            resource = Resource(name=key)
            resource.parent = parent
            database.session.add(resource)
        else:
            resource = persisted[key]

        if value:
            _persist_resources(value, resource)


class Ability(Model, ModelMixin):
    """An ability represents the capacity to perform an action (create, read,
    update, delete) on a resource/module/service of an application. In other
    words is an action/resource pair.

    This table is used to define those pairs, give them a name and a useful
    description.
    """

    __tablename__ = "ability"

    #: Foreign key. References to the column :attr:`~Action.id` of the table
    #: :class:`Action`.
    action_id = Column(Integer, ForeignKey("action.id"), primary_key=True)

    #: Foreign key. References to the column :attr:`~Resource.id` of the table
    #: :class:`Resource`.
    resource_id = Column(Integer, ForeignKey("resource.id"), primary_key=True)

    #: A name for the ability. For instance. Create Products.
    name = Column(String(80), nullable=False, unique=True)

    #: A long text that describes what this ability does.
    description = Column(Text)


def _persist_abilities():
    """This function generates **abilities** (action/resource pair) and inserts
    them into the database.

    Note that this doesn't commit the current session.
    """

    persisted = [(item.action_id, item.resource_id) for item in Ability.query.all()]

    actions = Action.query.all()
    resources = Resource.query.all()

    for resource in resources:
        for action in actions:
            if (action.id, resource.id) not in persisted:

                ability = Ability(
                    name=f"{action.name}:{resource.name}",
                    action_id=action.id,
                    resource_id=resource.id,
                )
                database.session.add(ability)


class Role(Model):
    """A Role is a set of **abilities** that can be assigned to organization
    members, for example, Seller, Cashier, Driver, Manager, etc.

    This table holds all roles of all organizations accounts, determining the
    organization that owns the role by the :class:`Org` identifier in the
    column :attr:`org_id`.

    Since the roles of all organizations reside in this table, the column
    :attr:`name` can have repeated values. But a role name must be unique
    in each organization.
    """

    __tablename__ = "role"

    #: Primary Key.
    id = Column(Integer, primary_key=True)

    #: A name for the paper, Cashier for example.
    name = Column(String(80), nullable=False)

    #: A long text that describes what this role does.
    description = Column(Text, nullable=False)

    #: The :attr:`~Org.id` of the organization account to which this role belongs.
    org_id = Column(Integer, ForeignKey("org.id"))

    #
    # -- Relationships --
    #

    abilities = relationship("RoleAbility", passive_deletes=True)

    __table_args__ = (
        # MemberRole has a composite foreign key referencing those columns.
        UniqueConstraint(id, org_id),
        # The name must be unique to each organization.
        UniqueConstraint(name, org_id),
    )


class RoleAbility(Model, ModelMixin):

    __tablename__ = "role_ability"

    role_id = Column(
        Integer, ForeignKey("role.id", ondelete="CASCADE"), primary_key=True
    )

    action_id = Column(Integer, nullable=False, primary_key=True)

    resource_id = Column(Integer, nullable=False, primary_key=True)

    __table_args__ = (
        ForeignKeyConstraint(
            [action_id, resource_id], [Ability.action_id, Ability.resource_id]
        ),
    )

    #
    # -- Relationships --
    #

    ability = relationship("Ability", uselist=False)

    role = relationship("Role", uselist=False)


class MemberRole(Model, ModelMixin):
    """All the roles that a user has in an organization.

    This table have two composite foreign keys:

    * (:attr:`org_id`, :attr:`user_id`) references to
      :class:`Membership` (:attr:`~Membership.org_id`, :attr:`~Membership.user_id`).
    * (:attr:`org_id`, :attr:`role_id`) references to
      :class:`Role` (:attr:`~Role.org_id`, :attr:`~Role.user_id`).

    Those two composite foreign keys ensure that the user to which a role is
    assigned indeed is a member of the organization.
    """

    __tablename__ = "member_role"

    #: Foreign key. Must be present in the tables :class:`Membership`
    #: and :class:`Role`.
    org_id = Column(Integer, nullable=False, primary_key=True)

    #: Foreign key with :attr:`~Membership.user_id` from the table
    #: :class:`Membership`.
    user_id = Column(Integer, nullable=False, primary_key=True)

    #: Foreign key. :attr:`Role.id` from the table :class:`Role`.
    role_id = Column(Integer, nullable=False, primary_key=True)

    #
    # -- Relationships --
    #

    role = relationship("Role", uselist=False)

    __table_args__ = (
        ForeignKeyConstraint(
            [user_id, org_id],
            [Membership.user_id, Membership.org_id],
            ondelete="CASCADE",
        ),
        ForeignKeyConstraint(
            [role_id, org_id], [Role.id, Role.org_id], ondelete="CASCADE"
        ),
    )


def get_member_privileges(org, user):
    member = Membership.query.filter_by(user_id=user.id, org_id=org.id).one()

    if member.is_owner:
        return {"org": ["manage"]}

    query = database.session.query

    # Get all roles from a specific member
    roles_subquery = (
        query(MemberRole.role_id)
        .filter_by(org_id=member.org_id, user_id=member.user_id)
        .subquery()
    )

    # Get abilities from the list of roles from the subquery above
    abilities_subquery = (
        query(RoleAbility).filter(RoleAbility.role_id.in_(roles_subquery)).subquery()
    )

    action = aliased(Action)
    resource = aliased(Resource)

    # Get all action and resource names from the abilities in the subquery above
    action_resource_query = (
        query(action.name.label("action"), resource.name.label("resource"))
        .select_from(abilities_subquery)
        .join(action, action.id == abilities_subquery.c.action_id)
        .join(resource, resource.id == abilities_subquery.c.resource_id)
    )

    abilities = action_resource_query.all()

    privileges = {}

    for ability in abilities:

        resource_name = ability.resource

        if resource_name not in privileges:
            privileges[resource_name] = []

        privileges[resource_name].append(ability.action)

    return privileges



import pytest
from saraki.auth import require_auth
from saraki.model import Resource, Action


@pytest.mark.usefixtures("ctx", "savepoint")
def test_init(app):
    @app.route("/routes")
    @require_auth("resource1", "action1")
    def route():
        pass

    app.init()

    assert Action.query.filter_by(name="action1").first()
    assert Resource.query.filter_by(name="resource1").first()
